# This is part of the MSS Python's module.
# Source: https://github.com/BoboTiG/python-mss.

import warnings
from typing import Any

from mss.base import MSS


def mss(**kwargs: Any) -> MSS:
    """Create an :class:`mss.MSS` instance for the current platform.

    .. deprecated:: 10.2.0
        Use :class:`mss.MSS` directly.
    """
    # TODO(jholveck): #493 Remove compatibility deprecation path once 10.x transition period ends.
    warnings.warn(
        "mss.mss is deprecated and will be removed in a future release; use mss.MSS instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return MSS(**kwargs)



"""XCB backend using MIT-SHM XShmGetImage with automatic fallback.

This is the fastest Linux backend available, and will work in most common
cases. However, it will not work over remote X connections, such as over ssh.

This implementation prefers shared-memory captures for performance and will
fall back to XGetImage when the MIT-SHM extension is unavailable or fails at
runtime. The fallback reason is exposed via ``performance_status`` to aid
debugging.

.. versionadded:: 10.2.0
"""

from __future__ import annotations

import enum
import os
from mmap import PROT_READ, mmap  # type: ignore[attr-defined]
from typing import TYPE_CHECKING, Any

from mss.exception import ScreenShotError
from mss.linux import xcb
from mss.linux.base import ALL_PLANES, MSSImplXCBBase
from mss.linux.xcbhelpers import LIB, XProtoError

if TYPE_CHECKING:
    from mss.models import Monitor

__all__ = ()


class ShmStatus(enum.Enum):
    """Availability of the MIT-SHM extension for this backend."""

    UNKNOWN = enum.auto()  # Constructor says SHM *should* work, but we haven't seen a real GetImage succeed yet.
    AVAILABLE = enum.auto()  # We've successfully used XShmGetImage at least once.
    UNAVAILABLE = enum.auto()  # We know SHM GetImage is unusable; always use XGetImage.


class MSSImplXShmGetImage(MSSImplXCBBase):
    """XCB backend using XShmGetImage with an automatic XGetImage fallback.

    .. seealso::
        :py:class:`mss.linux.base.MSSImplXCBBase`
            Lists constructor parameters.
    """

    def __init__(self, *, display: str | bytes | None = None, with_cursor: bool = False) -> None:
        super().__init__(display=display, with_cursor=with_cursor)

        # These are the objects we need to clean up when we shut down.  They are created in _setup_shm.
        self._memfd: int | None = None
        self._buf: mmap | None = None
        self._shmseg: xcb.ShmSeg | None = None

        # Rather than trying to track the shm_status, we may be able to raise an exception in __init__ if XShmGetImage
        # isn't available.  The factory in linux/__init__.py could then catch that and switch to XGetImage.
        # The conditions under which the attach will succeed but the xcb_shm_get_image will fail are extremely
        # rare, and I haven't yet found any that also will work with xcb_get_image.
        #: Whether we can use the MIT-SHM extensions for this connection.
        #: This will not be ``AVAILABLE`` until at least one capture has succeeded.
        #: It may be set to ``UNAVAILABLE`` sooner.
        self.shm_status: ShmStatus = self._setup_shm()

    def _shm_report_issue(self, msg: str, *args: Any) -> None:
        """Debugging hook for troubleshooting MIT-SHM issues.

        This will be called whenever MIT-SHM is disabled.  The optional
        arguments are not well-defined; exceptions are common.
        """
        full_msg = msg
        if args:
            full_msg += " | " + ", ".join(str(arg) for arg in args)
        self.performance_status.append(full_msg)

    def _setup_shm(self) -> ShmStatus:  # noqa: PLR0911
        assert self.conn is not None  # noqa: S101

        try:
            shm_ext_data = xcb.get_extension_data(self.conn, LIB.shm_id)
            if not shm_ext_data.present:
                self._shm_report_issue("MIT-SHM extension not present")
                return ShmStatus.UNAVAILABLE

            # We use the FD-based version of ShmGetImage, so we require the extension to be at least 1.2.
            shm_version_data = xcb.shm_query_version(self.conn)
            shm_version = (shm_version_data.major_version, shm_version_data.minor_version)
            if shm_version < (1, 2):
                self._shm_report_issue("MIT-SHM version too old", shm_version)
                return ShmStatus.UNAVAILABLE

            # We allocate something large enough for the root, so we don't have to reallocate each time the window is
            # resized.
            self._bufsize = self.pref_screen.width_in_pixels * self.pref_screen.height_in_pixels * 4

            if not hasattr(os, "memfd_create"):
                self._shm_report_issue("os.memfd_create not available")
                return ShmStatus.UNAVAILABLE
            try:
                self._memfd = os.memfd_create("mss-shm-buf", flags=os.MFD_CLOEXEC)  # type: ignore[attr-defined]
            except OSError as e:
                return self._shm_unavailable("memfd_create failed", e)
            os.ftruncate(self._memfd, self._bufsize)

            try:
                self._buf = mmap(self._memfd, self._bufsize, prot=PROT_READ)  # type: ignore[call-arg]
            except OSError as e:
                return self._shm_unavailable("mmap failed", e)
            self._shmseg = xcb.ShmSeg(xcb.generate_id(self.conn).value)
            try:
                # This will normally be what raises an exception if you're on a remote connection.
                # XCB will close _memfd, on success or on failure.
                try:
                    xcb.shm_attach_fd(self.conn, self._shmseg, self._memfd, read_only=False)
                finally:
                    self._memfd = None
            except xcb.XError as e:
                return self._shm_unavailable("Cannot attach MIT-SHM segment", e)
        except Exception:
            self._shutdown_shm()
            raise

        return ShmStatus.UNKNOWN

    def _shm_unavailable(self, msg: str, exc: Exception) -> ShmStatus:
        self._shm_report_issue(msg, exc)
        self._shutdown_shm()
        return ShmStatus.UNAVAILABLE

    def close(self) -> None:
        self._shutdown_shm()
        super().close()

    def _shutdown_shm(self) -> None:
        # It would be nice to also try to tell the server to detach the shmseg, but we might be in an error path
        # and don't know if that's possible.  It's not like we'll leak a lot of them on the same connection anyway.
        # This can be called in the path of partial initialization.
        if self._buf is not None:
            self._buf.close()
            self._buf = None
        if self._memfd is not None:
            os.close(self._memfd)
            self._memfd = None

    def _grab_xshmgetimage(self, monitor: Monitor) -> bytearray:
        if self.conn is None:
            msg = "Cannot take screenshot while the connection is closed"
            raise ScreenShotError(msg)
        assert self._buf is not None  # noqa: S101
        assert self._shmseg is not None  # noqa: S101

        required_size = monitor["width"] * monitor["height"] * 4
        if required_size > self._bufsize:
            # This is temporary.  The permanent fix will depend on how
            # issue https://github.com/BoboTiG/python-mss/issues/432 is resolved.
            msg = (
                "Requested capture size exceeds the allocated buffer. If you have resized the screen, "
                "please recreate your MSS object."
            )
            raise ScreenShotError(msg)

        img_reply = xcb.shm_get_image(
            self.conn,
            self.drawable,
            monitor["left"],
            monitor["top"],
            monitor["width"],
            monitor["height"],
            ALL_PLANES,
            xcb.ImageFormat.ZPixmap,
            self._shmseg,
            0,
        )

        if img_reply.depth != self.drawable_depth or img_reply.visual != self.drawable_visual_id:
            # This should never happen; a window can't change its visual.
            msg = (
                "Server returned an image with a depth or visual different than it initially reported: "
                f"expected {self.drawable_depth},{hex(self.drawable_visual_id.value)}, "
                f"got {img_reply.depth},{hex(img_reply.visual.value)}"
            )
            raise ScreenShotError(msg)

        # Snapshot the buffer into new bytearray.
        new_size = monitor["width"] * monitor["height"] * 4
        # Slicing the memoryview creates a new memoryview that points to the relevant subregion.  Making this and then
        # copying it into a fresh bytearray is much faster than slicing the mmap object.  Make sure we don't hold an
        # open memoryview if an exception happens, since that will prevent us from closing self._buf during the stack
        # unwind.
        with memoryview(self._buf) as img_mv:
            return bytearray(img_mv[:new_size])

    def grab(self, monitor: Monitor) -> bytearray:
        """Retrieve all pixels from a monitor. Pixels have to be RGBX."""
        if self.shm_status == ShmStatus.UNAVAILABLE:
            return super()._grab_xgetimage(monitor)

        # The usual path is just the next few lines.
        try:
            rv = self._grab_xshmgetimage(monitor)
            if self.shm_status != ShmStatus.AVAILABLE:
                self.shm_status = ShmStatus.AVAILABLE
                self.performance_status.append("MIT-SHM is working correctly.")
        except XProtoError as e:
            if self.shm_status != ShmStatus.UNKNOWN:
                # We know XShmGetImage works, because it worked earlier.  Reraise the error.
                raise

            # Should we engage the fallback path?  In almost all cases, if XShmGetImage failed at this stage (after
            # all our testing in __init__), XGetImage will also fail.  This could mean that the user sent an
            # out-of-bounds request.  In more exotic situations, some rare X servers disallow screen capture
            # altogether: security-hardened servers, for instance, or some XPrint servers.  But let's make sure, by
            # testing the same request through XGetImage.
            try:
                rv = super()._grab_xgetimage(monitor)
            except XProtoError:  # noqa: TRY203
                # The XGetImage also failed, so we don't know anything about whether XShmGetImage is usable.  Maybe
                # the user sent an out-of-bounds request.  Maybe it's a security-hardened server.  We're not sure what
                # the problem is.  So, if XGetImage failed, we re-raise that error (the one from XShmGetImage will be
                # attached as __context__), but we won't update the shm_status yet.  (Technically, our except:raise
                # clause here is redundant; it's just for clarity, to hold this comment.)
                raise

            # Using XShmGetImage failed, and using XGetImage worked.  Use XGetImage in the future.
            self._shm_report_issue("MIT-SHM GetImage failed", e)
            self.shm_status = ShmStatus.UNAVAILABLE
            self._shutdown_shm()

        return rv



"""Microsoft Windows backend dispatcher for MSS screenshot implementations."""

from __future__ import annotations

import warnings
from typing import Any

from mss.base import MSS as _MSS
from mss.base import MSSImplementation
from mss.exception import ScreenShotError

__all__ = ["MSS"]

BACKENDS = ["default", "gdi"]


class MSS(_MSS):
    """Deprecated Microsoft Windows compatibility constructor.

    Use :class:`mss.MSS` instead.
    """

    def __init__(self, /, **kwargs: Any) -> None:
        # TODO(jholveck): #493 Remove compatibility constructor after 10.x transition period.
        warnings.warn(
            "mss.windows.MSS is deprecated and will be removed in 11.0; use mss.MSS instead",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(**kwargs)


def choose_impl(backend: str = "default", **kwargs: Any) -> MSSImplementation:
    """Return a backend-specific MSS implementation for Microsoft Windows.

    Selects and instantiates the appropriate Windows backend based on the
    ``backend`` parameter.

    :param backend: Backend selector. Valid values:

        - ``"default"`` or ``"gdi"`` (default): GDI-based backend using
          ``BitBlt`` and ``CreateDIBSection`` for direct memory access to pixel
          data.

    :param kwargs: Additional keyword arguments passed to the backend class.
    :returns: An MSS backend implementation.

    .. versionadded:: 10.2.0 Prior to this version, this didn't exist:
          Windows had a single implementation selected through
          :class:`mss.windows.MSS`.
    """
    backend = backend.lower()
    if backend in {"default", "gdi"}:
        from mss.windows.gdi import MSSImplGdi  # noqa: PLC0415

        return MSSImplGdi(**kwargs)
    assert backend not in BACKENDS  # noqa: S101
    msg = f"Backend {backend!r} not (yet?) implemented."
    raise ScreenShotError(msg)



from __future__ import annotations

import ctypes.util
from contextlib import suppress
from copy import copy
from ctypes import (
    CDLL,
    POINTER,
    Array,
    Structure,
    _Pointer,
    addressof,
    c_char,
    c_char_p,
    c_int,
    c_uint,
    c_uint8,
    c_uint16,
    c_uint32,
    c_void_p,
    cast,
    cdll,
)
from threading import RLock
from typing import TYPE_CHECKING
from weakref import finalize

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from typing import Any

from mss.exception import ScreenShotError

# A quick refresher on why this module spends so much effort on object lifetimes, and how the pieces fit together:
#
# 1. Shape of XCB replies.
#    Each reply that comes back from libxcb is one contiguous allocation that looks like:
#       [fixed-size header][optional padding][embedded arrays/lists]
#    The protocol spec describes where those trailing lists live, but callers are not expected to compute offsets by
#    hand.  Instead, XCB exposes helper functions such as `xcb_setup_pixmap_formats` (returns pointer + length for a
#    fixed-size array) or iterator factories for nested variable-length data.  As long as the original reply is still
#    allocated, all of the derived pointers remain valid.
#
# 2. What ctypes does (and does not) track automatically.
#    When user code reads `my_struct.foo`, ctypes returns another ctypes object that still refers to memory owned by
#    `my_struct`; it does not copy the value.  To keep that relationship alive, ctypes silently sets `_b_base_` on the
#    derived object so the garbage collector knows that `my_struct` must stay around.  This mechanism only works when
#    ctypes itself materializes the derived object.
#
# 3. Why XCB accessors break that safety net.
#    The XCB helpers we need - `xcb_setup_pixmap_formats`, `xcb_randr_get_screen_resources_crtcs`, etc. - return raw C
#    pointers.  ctypes happily converts them to Python objects, but because the conversion went through a plain C
#    call, `_b_base_` never gets filled in.  The GC no longer realizes that the derived array depends on the reply, so
#    once every direct reference to the reply drops, libc is free to `free()` the allocation.  Any later access
#    through the derived pointer becomes undefined behaviour.
#
# 4. How this module keeps everything safe.
#    After every call into an XCB accessor we immediately call `depends_on(child, parent)`.  That helper installs a
#    finalizer on `child` whose only job is to keep a reference to `parent`.  No extra work is performed; the callback
#    holding the reference is enough to keep the reply alive until the child objects disappear.  Separately, when we
#    first receive the reply, we register another finalizer that hands the pointer back to libc once *all* dependants
#    have been collected.  As a result, higher-level code can treat these helper functions just like the XCB C API:
#    grab the array you need, keep it as long as you like, and trust that it stays valid.


def depends_on(subobject: Any, superobject: Any) -> None:
    """Make sure that superobject is not GC'd before subobject.

    In XCB, a structure often is allocated with additional trailing
    data following it, with special accessors to get pointers to that
    extra data.

    In ctypes, if you access a structure field, a pointer value, etc.,
    then the outer object won't be garbage collected until after the
    inner object.  (This uses the ctypes _b_base_ mechanism.)

    However, when using the XCB accessor functions, you don't get that
    guarantee automatically.  Once all references to the outer
    structure have dropped, then we will free the memory for it (the
    response structures XCB returns have to be freed by us), including
    the trailing data.  If there are live references to the trailing
    data, then those will become invalid.

    To prevent this, we use depends_on to make sure that the
    outer structure is not released before all the references to the
    inner objects have been cleared.
    """
    # The implementation is quite simple.  We create a finalizer on the inner object, with a callback that references
    # the outer object.  That ensures that there are live references to the outer object until the references to the
    # inner object have been gc'd.  We can't just create a ref, though; it seems that their callbacks will only run if
    # the ref itself is still referenced.  We need the extra machinery that finalize provides, which uses an internal
    # registry to keep the refs alive.
    finalize(subobject, id, superobject)


#### XCB basic structures


class Connection(Structure):
    pass  # Opaque


class XID(c_uint32):
    def __eq__(self, other: object) -> bool:
        return self.value == other.value if isinstance(other, XID) else NotImplemented

    def __hash__(self) -> int:
        return hash(self.value)


class GenericErrorStructure(Structure):
    # The XCB name in C is xcb_generic_error.  It is named differently here to make it clear that this is not an
    # exception class, since in Python, those traditionally end in ...Error.
    _fields_ = (
        ("response_type", c_uint8),
        ("error_code", c_uint8),
        ("sequence", c_uint16),
        ("resource_id", c_uint32),
        ("minor_code", c_uint16),
        ("major_code", c_uint8),
        ("pad0", c_uint8),
        ("pad", c_uint32 * 5),
        ("full_sequence", c_uint32),
    )


# We special-case InternAtom for convenience.
class InternAtomReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("pad0", c_uint8 * 1),
        ("sequence", c_uint16),
        ("length", c_uint32),
        # This is actually an Atom, not a raw XID, but we handle the type conversion in intern_atom.
        ("atom", XID),
    )


#### Request / response handling
#
# The following recaps a lot of what's in the xcb-requests(3) man page, with a few notes about what we're doing in
# this library.
#
# In XCB, when you send a request to the server, the function returns immediately.  You don't get back the server's
# reply; you get back a "cookie".  (This just holds the sequence number of the request.)  Later, you can use that
# cookie to get the reply or error back.
#
# This lets you fire off requests in rapid succession, and then afterwards check the results.  It also lets you do
# other work (like process a screenshot) while a request is in flight (like getting the next screenshot).  This is
# asynchronous processing, and is great for performance.
#
# In this program, we currently don't try to do anything asynchronously, although the design doesn't preclude it.
# (You'd add a synchronous=False flag to the entrypoint wrappers below, and not call .check / .reply, but rather just
# return the cookie.)
#
# XCB has two types of requests.  Void requests don't return anything from the server.  These are things like "create
# a window".  The typed requests do request information from the server.  These are things like "get a window's size".
#
# Void requests all return the same type of cookie.  The only thing you can do with the cookie is check to see if you
# got an error.
#
# Typed requests return a call-specific cookie with the same structure.  They are call-specific so they can be
# type-checked.  (This is the case in both XCB C and in this library.)
#
# XCB has a concept of "checked" or "unchecked" request functions.  By default, void requests are unchecked.  For an
# unchecked function, XCB doesn't do anything to let you know that the request completed successfully.  If there's an
# error, then you need to handle it in your main loop, as a regular event.  We always use the checked versions
# instead, so that we can raise an exception at the right place in the code.
#
# Similarly, typed requests default to checked, but have unchecked versions.  That's just to align their error
# handling with the unchecked void functions; you always need to do something with the cookie so you can get the
# response.
#
# As mentioned, we always use the checked requests; that's unlikely to change, since error-checking with unchecked
# requests requires control of the event loop.
#
# Below are wrappers that set up the request / response functions in ctypes, and define the cookie types to do error
# handling.


class XError(ScreenShotError):
    """Base exception class for anything related to X11.

    This is not prefixed with Xcb to prevent confusion with the XCB
    error structures.
    """


class XProtoError(XError):
    """Exception indicating server-reported errors."""

    def __init__(self, xcb_conn: Connection, xcb_err: GenericErrorStructure) -> None:
        if isinstance(xcb_err, _Pointer):
            xcb_err = xcb_err.contents
        assert isinstance(xcb_err, GenericErrorStructure)  # noqa: S101

        details = {
            "error_code": xcb_err.error_code,
            "sequence": xcb_err.sequence,
            "resource_id": xcb_err.resource_id,
            "minor_code": xcb_err.minor_code,
            "major_code": xcb_err.major_code,
            "full_sequence": xcb_err.full_sequence,
        }

        # xcb-errors is a library to get descriptive error strings, instead of reporting the raw codes.  This is not
        # installed by default on most systems, but is quite helpful for developers.  We use it if it exists, but
        # don't force the matter.  We can't delay this lookup until we format the error message, since the XCB
        # connection may be gone by then.
        if LIB.errors:
            # We don't try to reuse the error context, since it's per-connection, and probably will only be used once.
            ctx = POINTER(XcbErrorsContext)()
            ctx_new_setup = LIB.errors.xcb_errors_context_new(xcb_conn, ctx)
            if ctx_new_setup == 0:
                try:
                    # Some of these may return NULL, but some are guaranteed.
                    ext_name = POINTER(c_char_p)()
                    error_name = LIB.errors.xcb_errors_get_name_for_error(ctx, xcb_err.error_code, ext_name)
                    details["error"] = error_name.decode("ascii", errors="replace")
                    if ext_name:
                        ext_name_str = ext_name.contents.value
                        # I'm pretty sure it'll always be populated if ext_name is set, but...
                        if ext_name_str is not None:
                            details["extension"] = ext_name_str.decode("ascii", errors="replace")
                    major_name = LIB.errors.xcb_errors_get_name_for_major_code(ctx, xcb_err.major_code)
                    details["major_name"] = major_name.decode("ascii", errors="replace")
                    minor_name = LIB.errors.xcb_errors_get_name_for_minor_code(
                        ctx, xcb_err.major_code, xcb_err.minor_code
                    )
                    if minor_name:
                        details["minor_name"] = minor_name.decode("ascii", errors="replace")
                finally:
                    LIB.errors.xcb_errors_context_free(ctx)

        super().__init__("X11 Protocol Error", details=details)

    def __str__(self) -> str:
        msg = super().__str__()
        details = self.details
        error_desc = f"{details['error_code']} ({details['error']})" if "error" in details else details["error_code"]
        major_desc = (
            f"{details['major_code']} ({details['major_name']})" if "major_name" in details else details["major_code"]
        )
        minor_desc = (
            f"{details['minor_code']} ({details['minor_name']})" if "minor_name" in details else details["minor_code"]
        )
        ext_desc = f"\n  Extension:  {details['extension']}" if "extension" in details else ""
        msg += (
            f"\nX Error of failed request:  {error_desc}"
            f"\n  Major opcode of failed request:  {major_desc}{ext_desc}"
            + (f"\n  Minor opcode of failed request:  {minor_desc}" if details["minor_code"] != 0 else "")
            + f"\n  Resource id in failed request:  {details['resource_id']}"
            f"\n  Serial number of failed request:  {details['full_sequence']}"
        )
        return msg


class CookieBase(Structure):
    """Generic XCB cookie.

    XCB does not export this as a base type.  However, all XCB cookies
    have the same structure, so this encompasses the common structure
    in Python.
    """

    # It's possible to add a finalizer that will raise an exception if a cookie is garbage collected without being
    # disposed of (through discard, check, or reply).  If we ever start using asynchronous requests, then that would
    # be good to add.  But for now, we can trust the wrapper functions to manage the cookies correctly, without the
    # extra overhead of these finalizers.

    _fields_ = (("sequence", c_uint),)

    def discard(self, xcb_conn: Connection) -> None:
        """Free memory associated with this request, and ignore errors."""
        LIB.xcb.xcb_discard_reply(xcb_conn, self.sequence)


class VoidCookie(CookieBase):
    """XCB cookie for requests with no responses.

    This corresponds to xcb_void_cookie_t.
    """

    def check(self, xcb_conn: Connection) -> None:
        """Verify that the function completed successfully.

        This will raise an exception if there is an error.
        """
        err_p = LIB.xcb.xcb_request_check(xcb_conn, self)
        if not err_p:
            return
        err = copy(err_p.contents)
        LIB.c.free(err_p)
        raise XProtoError(xcb_conn, err)


class ReplyCookieBase(CookieBase):
    _xcb_reply_func = None

    def reply(self, xcb_conn: Connection) -> Structure:
        """Wait for and return the server's response.

        The response will be freed (with libc's free) when it, and its
        descendents, are no longer referenced.

        If the server indicates an error, an exception is raised
        instead.
        """
        err_p = POINTER(GenericErrorStructure)()
        assert self._xcb_reply_func is not None  # noqa: S101
        reply_p = self._xcb_reply_func(xcb_conn, self, err_p)
        if err_p:
            # I think this is always NULL, but we can free it.
            if reply_p:
                LIB.c.free(reply_p)
            # Copying the error structure is cheap, and makes memory management easier.
            err_copy = copy(err_p.contents)
            LIB.c.free(err_p)
            raise XProtoError(xcb_conn, err_copy)
        assert reply_p  # noqa: S101

        # It's not known, at this point, how long the reply structure actually is: there may be trailing data that
        # needs to be processed and then freed.  We have to set a finalizer on the reply, so it can be freed when
        # Python is done with it.  The whole dependency tree, though, leads back to this object and its finalizer.
        # Importantly, reply_void_p does not carry a reference (direct or indirect) to reply_p; that would prevent
        # it from ever being freed.
        reply_void_p = c_void_p(addressof(reply_p.contents))
        finalizer = finalize(reply_p, LIB.c.free, reply_void_p)
        finalizer.atexit = False
        return reply_p.contents


def initialize_xcb_typed_func(lib: CDLL, name: str, request_argtypes: list, reply_struct: type) -> None:
    """Set up ctypes for a response-returning XCB function.

    This is only applicable to checked (the default) variants of
    functions that have a response type.

    This arranges for the ctypes function to take the given argtypes.
    The ctypes function will then return an XcbTypedCookie (rather,
    a function-specific subclass of it).  That can be used to call the
    XCB xcb_blahblah_reply function to check for errors and return the
    server's response.
    """

    base_name = name
    title_name = base_name.title().replace("_", "")
    request_func = getattr(lib, name)
    reply_func = getattr(lib, f"{name}_reply")
    # The cookie type isn't used outside this function, so we can just declare it here implicitly.
    cookie_type = type(f"{title_name}Cookie", (ReplyCookieBase,), {"_xcb_reply_func": reply_func})
    request_func.argtypes = request_argtypes
    request_func.restype = cookie_type
    reply_func.argtypes = [POINTER(Connection), cookie_type, POINTER(POINTER(GenericErrorStructure))]
    reply_func.restype = POINTER(reply_struct)


### XCB types


class XcbExtension(Structure):
    _fields_ = (("name", c_char_p), ("global_id", c_int))


class XcbErrorsContext(Structure):
    """A context for using libxcb-errors.

    Create a context with xcb_errors_context_new() and destroy it with
    xcb_errors_context_free(). Except for xcb_errors_context_free(),
    all functions in libxcb-errors are thread-safe and can be called
    from multiple threads at the same time, even on the same context.
    """


#### Types for special-cased functions


class QueryExtensionReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("pad0", c_uint8),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("present", c_uint8),
        ("major_opcode", c_uint8),
        ("first_event", c_uint8),
        ("first_error", c_uint8),
    )


#### XCB libraries singleton


class LibContainer:
    """Container for XCB-related libraries.

    There is one instance exposed as the xcb.LIB global.

    You can access libxcb.so as xcb.LIB.xcb, libc as xcb.LIB.c, etc.
    These are not set up until initialize() is called.  It is safe to
    call initialize() multiple times.

    Library accesses through this container return the ctypes CDLL
    object.  There are no smart wrappers (although the return types are
    the cookie classes defined above).  In other words, if you're
    accessing xcb.LIB.xcb.xcb_foo, then you need to handle the .reply()
    calls and such yourself.  If you're accessing the wrapper functions
    in the xcb module xcb.foo, then it will take care of that for you.
    """

    _EXPOSED_NAMES = frozenset(
        {"c", "xcb", "randr", "randr_id", "render", "render_id", "xfixes", "xfixes_id", "errors"}
    )

    def __init__(self) -> None:
        self._lock = RLock()
        self._initializing = False
        self.initialized = False

    def reset(self) -> None:
        with self._lock:
            if self._initializing:
                msg = "Cannot reset during initialization"
                raise RuntimeError(msg)
            self.initialized = False
            for name in self._EXPOSED_NAMES:
                with suppress(AttributeError):
                    delattr(self, name)

    def initialize(self, callbacks: Iterable[Callable[[], None]] = frozenset()) -> None:  # noqa: PLR0915
        # We'll need a couple of generated types, but we have to load them late, since xcbgen requires this library.
        from mss.linux.xcbgen import Setup  # noqa: PLC0415

        with self._lock:
            if self.initialized:
                # Something else initialized this object while we were waiting for the lock.
                return

            if self._initializing:
                msg = "Cannot load during initialization"
                raise RuntimeError(msg)

            try:
                self._initializing = True

                # We don't use the cached versions that ctypes.cdll exposes as attributes, since other libraries may be
                # doing their own things with these.

                # We use the libc that the current process has loaded, to make sure we get the right version of free().
                # ctypes doesn't document that None is valid as the argument to LoadLibrary, but it does the same thing
                # as a NULL argument to dlopen: it returns the current process and its loaded libraries.  This includes
                # libc.
                self.c = cdll.LoadLibrary(None)  # type: ignore[arg-type]
                self.c.free.argtypes = [c_void_p]
                self.c.free.restype = None

                libxcb_so = ctypes.util.find_library("xcb")
                if libxcb_so is None:
                    msg = "Library libxcb.so not found"
                    raise ScreenShotError(msg)
                self.xcb = cdll.LoadLibrary(libxcb_so)

                self.xcb.xcb_request_check.argtypes = [POINTER(Connection), VoidCookie]
                self.xcb.xcb_request_check.restype = POINTER(GenericErrorStructure)
                self.xcb.xcb_discard_reply.argtypes = [POINTER(Connection), c_uint]
                self.xcb.xcb_discard_reply.restype = None
                self.xcb.xcb_get_extension_data.argtypes = [POINTER(Connection), POINTER(XcbExtension)]
                self.xcb.xcb_get_extension_data.restype = POINTER(QueryExtensionReply)
                self.xcb.xcb_prefetch_extension_data.argtypes = [POINTER(Connection), POINTER(XcbExtension)]
                self.xcb.xcb_prefetch_extension_data.restype = None
                self.xcb.xcb_generate_id.argtypes = [POINTER(Connection)]
                self.xcb.xcb_generate_id.restype = XID
                self.xcb.xcb_get_setup.argtypes = [POINTER(Connection)]
                self.xcb.xcb_get_setup.restype = POINTER(Setup)
                self.xcb.xcb_connection_has_error.argtypes = [POINTER(Connection)]
                self.xcb.xcb_connection_has_error.restype = c_int
                self.xcb.xcb_connect.argtypes = [c_char_p, POINTER(c_int)]
                self.xcb.xcb_connect.restype = POINTER(Connection)
                self.xcb.xcb_disconnect.argtypes = [POINTER(Connection)]
                self.xcb.xcb_disconnect.restype = None

                # We special-case InternAtom for convenience.
                initialize_xcb_typed_func(
                    LIB.xcb,
                    "xcb_intern_atom",
                    [POINTER(Connection), c_uint8, c_uint16, POINTER(c_char)],
                    InternAtomReply,
                )

                libxcb_randr_so = ctypes.util.find_library("xcb-randr")
                if libxcb_randr_so is None:
                    msg = "Library libxcb-randr.so not found"
                    raise ScreenShotError(msg)
                self.randr = cdll.LoadLibrary(libxcb_randr_so)
                self.randr_id = XcbExtension.in_dll(self.randr, "xcb_randr_id")

                libxcb_render_so = ctypes.util.find_library("xcb-render")
                if libxcb_render_so is None:
                    msg = "Library libxcb-render.so not found"
                    raise ScreenShotError(msg)
                self.render = cdll.LoadLibrary(libxcb_render_so)
                self.render_id = XcbExtension.in_dll(self.render, "xcb_render_id")

                libxcb_shm_so = ctypes.util.find_library("xcb-shm")
                if libxcb_shm_so is None:
                    msg = "Library libxcb-shm.so not found"
                    raise ScreenShotError(msg)
                self.shm = cdll.LoadLibrary(libxcb_shm_so)
                self.shm_id = XcbExtension.in_dll(self.shm, "xcb_shm_id")

                libxcb_xfixes_so = ctypes.util.find_library("xcb-xfixes")
                if libxcb_xfixes_so is None:
                    msg = "Library libxcb-xfixes.so not found"
                    raise ScreenShotError(msg)
                self.xfixes = cdll.LoadLibrary(libxcb_xfixes_so)
                self.xfixes_id = XcbExtension.in_dll(self.xfixes, "xcb_xfixes_id")

                # xcb_errors is an optional library, mostly only useful to developers.  We use the qualified .so name,
                # since it's subject to change incompatibly.
                try:
                    self.errors: CDLL | None = cdll.LoadLibrary("libxcb-errors.so.0")
                except Exception:  # noqa: BLE001
                    self.errors = None
                else:
                    self.errors.xcb_errors_context_new.argtypes = [
                        POINTER(Connection),
                        POINTER(POINTER(XcbErrorsContext)),
                    ]
                    self.errors.xcb_errors_context_new.restype = c_int
                    self.errors.xcb_errors_context_free.argtypes = [POINTER(XcbErrorsContext)]
                    self.errors.xcb_errors_context_free.restype = None
                    self.errors.xcb_errors_get_name_for_major_code.argtypes = [POINTER(XcbErrorsContext), c_uint8]
                    self.errors.xcb_errors_get_name_for_major_code.restype = c_char_p
                    self.errors.xcb_errors_get_name_for_minor_code.argtypes = [
                        POINTER(XcbErrorsContext),
                        c_uint8,
                        c_uint16,
                    ]
                    self.errors.xcb_errors_get_name_for_minor_code.restype = c_char_p
                    self.errors.xcb_errors_get_name_for_error.argtypes = [
                        POINTER(XcbErrorsContext),
                        c_uint8,
                        POINTER(c_char_p),
                    ]
                    self.errors.xcb_errors_get_name_for_error.restype = c_char_p

                for x in callbacks:
                    x()

            finally:
                self._initializing = False

            self.initialized = True


LIB = LibContainer()


#### Trailing data accessors
#
# In X11, many replies have the header (the *Reply structures defined above), plus some variable-length data after it.
# For instance, XcbScreen includes a list of XcbDepth structures.
#
# These mostly follow two patterns.
#
# For objects with a constant size, we get a pointer and length (count), cast to an array, and return the array
# contents.  (This doesn't involve copying any data.)
#
# For objects with a variable size, we use the XCB-provided iterator protocol to iterate over them, and return a
# Python list.  (This also doesn't copy any data, but does construct a list.)  To continue the example of how
# XcbScreen includes a list of XcbDepth structures: a full XcbDepth is variable-length because it has a variable
# number of visuals attached to it.
#
# These lists with variable element sizes follow a standard pattern:
#
# * There is an iterator class (such as XcbScreenIterator), based on the type you're iterating over.  This defines a
#   data pointer to point to the current object, and a rem counter indicating the remaining number of objects.
# * There is a function to advance the iterator (such as xcb_screen_next), based on the type of iterator being
#   advanced.
# * There is an initializer function (such as xcb_setup_roots_iterator) that takes the container (XcbSetup), and
#   returns an iterator (XcbScreenIterator) pointing to the first object in the list.  (This iterator is returned by
#   value, so Python can free it normally.)
#
# The returned structures are actually part of the allocation of the parent pointer: the POINTER(XcbScreen) objects
# point to objects that were allocated along with the XcbSetup that we got them from.  That means that it is very
# important that the XcbSetup not be freed until the pointers that point into it are freed.


### Iteration utility primitives


def list_from_xcb(iterator_factory: Callable, next_func: Callable, parent: Structure | _Pointer) -> list:
    iterator = iterator_factory(parent)
    items: list = []
    while iterator.rem != 0:
        current = iterator.data.contents
        # Keep the parent reply alive until consumers drop this entry.
        depends_on(current, parent)
        items.append(current)
        next_func(iterator)
    return items


def array_from_xcb(pointer_func: Callable, length_func: Callable, parent: Structure | _Pointer) -> Array:
    pointer = pointer_func(parent)
    length = length_func(parent)
    if length and not pointer:
        msg = "XCB returned a NULL pointer for non-zero data length"
        raise ScreenShotError(msg)
    array_ptr = cast(pointer, POINTER(pointer._type_ * length))
    array = array_ptr.contents
    depends_on(array, parent)
    return array



# Auto-generated by gen_xcb_to_py.py - do not edit manually.

# Since many of the generated functions have many parameters, we disable the pylint warning about too many arguments.
# ruff: noqa: PLR0913

from __future__ import annotations

from ctypes import (
    POINTER,
    Array,
    Structure,
    _Pointer,
    c_char,
    c_int,
    c_int16,
    c_uint8,
    c_uint16,
    c_uint32,
)
from enum import IntEnum

from mss.linux.xcbhelpers import (
    LIB,
    XID,
    Connection,
    VoidCookie,
    array_from_xcb,
    initialize_xcb_typed_func,
    list_from_xcb,
)

RANDR_MAJOR_VERSION = 1
RANDR_MINOR_VERSION = 6
RENDER_MAJOR_VERSION = 0
RENDER_MINOR_VERSION = 11
SHM_MAJOR_VERSION = 1
SHM_MINOR_VERSION = 2
XFIXES_MAJOR_VERSION = 6
XFIXES_MINOR_VERSION = 0

# Enum classes


class RandrConnection(IntEnum):
    Connected = 0
    Disconnected = 1
    Unknown = 2


class RandrSetConfig(IntEnum):
    Success = 0
    InvalidConfigTime = 1
    InvalidTime = 2
    Failed = 3


class RenderPictType(IntEnum):
    Indexed = 0
    Direct = 1


class RenderSubPixel(IntEnum):
    Unknown = 0
    HorizontalRGB = 1
    HorizontalBGR = 2
    VerticalRGB = 3
    VerticalBGR = 4
    None_ = 5


class BackingStore(IntEnum):
    NotUseful = 0
    WhenMapped = 1
    Always = 2


class ImageFormat(IntEnum):
    XYBitmap = 0
    XYPixmap = 1
    ZPixmap = 2


class ImageOrder(IntEnum):
    LSBFirst = 0
    MSBFirst = 1


class VisualClass(IntEnum):
    StaticGray = 0
    GrayScale = 1
    StaticColor = 2
    PseudoColor = 3
    TrueColor = 4
    DirectColor = 5


# Generated ctypes structures


class Drawable(XID):
    pass


class Keycode(c_uint8):
    def __eq__(self, other: object) -> bool:
        if isinstance(other, Keycode):
            return self.value == other.value
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.value)


class Format(Structure):
    _fields_ = (
        ("depth", c_uint8),
        ("bits_per_pixel", c_uint8),
        ("scanline_pad", c_uint8),
        ("pad0", c_uint8 * 5),
    )


class Window(Drawable):
    pass


class Colormap(XID):
    pass


class Visualid(c_uint32):
    def __eq__(self, other: object) -> bool:
        if isinstance(other, Visualid):
            return self.value == other.value
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.value)


class Visualtype(Structure):
    _fields_ = (
        ("visual_id", Visualid),
        ("class_", c_uint8),
        ("bits_per_rgb_value", c_uint8),
        ("colormap_entries", c_uint16),
        ("red_mask", c_uint32),
        ("green_mask", c_uint32),
        ("blue_mask", c_uint32),
        ("pad0", c_uint8 * 4),
    )


class Depth(Structure):
    _fields_ = (
        ("depth", c_uint8),
        ("pad0", c_uint8 * 1),
        ("visuals_len", c_uint16),
        ("pad1", c_uint8 * 4),
    )


class DepthIterator(Structure):
    _fields_ = (("data", POINTER(Depth)), ("rem", c_int), ("index", c_int))


class Screen(Structure):
    _fields_ = (
        ("root", Window),
        ("default_colormap", Colormap),
        ("white_pixel", c_uint32),
        ("black_pixel", c_uint32),
        ("current_input_masks", c_uint32),
        ("width_in_pixels", c_uint16),
        ("height_in_pixels", c_uint16),
        ("width_in_millimeters", c_uint16),
        ("height_in_millimeters", c_uint16),
        ("min_installed_maps", c_uint16),
        ("max_installed_maps", c_uint16),
        ("root_visual", Visualid),
        ("backing_stores", c_uint8),
        ("save_unders", c_uint8),
        ("root_depth", c_uint8),
        ("allowed_depths_len", c_uint8),
    )


class ScreenIterator(Structure):
    _fields_ = (("data", POINTER(Screen)), ("rem", c_int), ("index", c_int))


class Setup(Structure):
    _fields_ = (
        ("status", c_uint8),
        ("pad0", c_uint8 * 1),
        ("protocol_major_version", c_uint16),
        ("protocol_minor_version", c_uint16),
        ("length", c_uint16),
        ("release_number", c_uint32),
        ("resource_id_base", c_uint32),
        ("resource_id_mask", c_uint32),
        ("motion_buffer_size", c_uint32),
        ("vendor_len", c_uint16),
        ("maximum_request_length", c_uint16),
        ("roots_len", c_uint8),
        ("pixmap_formats_len", c_uint8),
        ("image_byte_order", c_uint8),
        ("bitmap_format_bit_order", c_uint8),
        ("bitmap_format_scanline_unit", c_uint8),
        ("bitmap_format_scanline_pad", c_uint8),
        ("min_keycode", Keycode),
        ("max_keycode", Keycode),
        ("pad1", c_uint8 * 4),
    )


class SetupIterator(Structure):
    _fields_ = (("data", POINTER(Setup)), ("rem", c_int), ("index", c_int))


class Pixmap(Drawable):
    pass


class GetGeometryReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("depth", c_uint8),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("root", Window),
        ("x", c_int16),
        ("y", c_int16),
        ("width", c_uint16),
        ("height", c_uint16),
        ("border_width", c_uint16),
        ("pad0", c_uint8 * 10),
    )


class GetImageReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("depth", c_uint8),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("visual", Visualid),
        ("pad0", c_uint8 * 20),
    )


class Atom(XID):
    pass


class GetPropertyReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("format_", c_uint8),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("type_", Atom),
        ("bytes_after", c_uint32),
        ("value_len", c_uint32),
        ("pad0", c_uint8 * 12),
    )


class RandrQueryVersionReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("pad0", c_uint8 * 1),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("major_version", c_uint32),
        ("minor_version", c_uint32),
        ("pad1", c_uint8 * 16),
    )


class Timestamp(c_uint32):
    def __eq__(self, other: object) -> bool:
        if isinstance(other, Timestamp):
            return self.value == other.value
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.value)


class RandrCrtc(XID):
    pass


class RandrOutput(XID):
    pass


class RandrModeInfo(Structure):
    _fields_ = (
        ("id_", c_uint32),
        ("width", c_uint16),
        ("height", c_uint16),
        ("dot_clock", c_uint32),
        ("hsync_start", c_uint16),
        ("hsync_end", c_uint16),
        ("htotal", c_uint16),
        ("hskew", c_uint16),
        ("vsync_start", c_uint16),
        ("vsync_end", c_uint16),
        ("vtotal", c_uint16),
        ("name_len", c_uint16),
        ("mode_flags", c_uint32),
    )


class RandrGetScreenResourcesReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("pad0", c_uint8 * 1),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("timestamp", Timestamp),
        ("config_timestamp", Timestamp),
        ("num_crtcs", c_uint16),
        ("num_outputs", c_uint16),
        ("num_modes", c_uint16),
        ("names_len", c_uint16),
        ("pad1", c_uint8 * 8),
    )


class RandrGetScreenResourcesCurrentReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("pad0", c_uint8 * 1),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("timestamp", Timestamp),
        ("config_timestamp", Timestamp),
        ("num_crtcs", c_uint16),
        ("num_outputs", c_uint16),
        ("num_modes", c_uint16),
        ("names_len", c_uint16),
        ("pad1", c_uint8 * 8),
    )


class RandrMode(XID):
    pass


class RandrGetCrtcInfoReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("status", c_uint8),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("timestamp", Timestamp),
        ("x", c_int16),
        ("y", c_int16),
        ("width", c_uint16),
        ("height", c_uint16),
        ("mode", RandrMode),
        ("rotation", c_uint16),
        ("rotations", c_uint16),
        ("num_outputs", c_uint16),
        ("num_possible_outputs", c_uint16),
    )


class RandrGetOutputInfoReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("status", c_uint8),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("timestamp", Timestamp),
        ("crtc", RandrCrtc),
        ("mm_width", c_uint32),
        ("mm_height", c_uint32),
        ("connection", c_uint8),
        ("subpixel_order", c_uint8),
        ("num_crtcs", c_uint16),
        ("num_modes", c_uint16),
        ("num_preferred", c_uint16),
        ("num_clones", c_uint16),
        ("name_len", c_uint16),
    )


class RandrGetOutputPrimaryReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("pad0", c_uint8 * 1),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("output", RandrOutput),
    )


class RandrGetOutputPropertyReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("format_", c_uint8),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("type_", Atom),
        ("bytes_after", c_uint32),
        ("num_items", c_uint32),
        ("pad0", c_uint8 * 12),
    )


class RandrMonitorInfo(Structure):
    _fields_ = (
        ("name", Atom),
        ("primary", c_uint8),
        ("automatic", c_uint8),
        ("nOutput", c_uint16),
        ("x", c_int16),
        ("y", c_int16),
        ("width", c_uint16),
        ("height", c_uint16),
        ("width_in_millimeters", c_uint32),
        ("height_in_millimeters", c_uint32),
    )


class RandrMonitorInfoIterator(Structure):
    _fields_ = (("data", POINTER(RandrMonitorInfo)), ("rem", c_int), ("index", c_int))


class RandrGetMonitorsReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("pad0", c_uint8 * 1),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("timestamp", Timestamp),
        ("nMonitors", c_uint32),
        ("nOutputs", c_uint32),
        ("pad1", c_uint8 * 12),
    )


class RenderQueryVersionReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("pad0", c_uint8 * 1),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("major_version", c_uint32),
        ("minor_version", c_uint32),
        ("pad1", c_uint8 * 16),
    )


class RenderPictformat(XID):
    pass


class RenderDirectformat(Structure):
    _fields_ = (
        ("red_shift", c_uint16),
        ("red_mask", c_uint16),
        ("green_shift", c_uint16),
        ("green_mask", c_uint16),
        ("blue_shift", c_uint16),
        ("blue_mask", c_uint16),
        ("alpha_shift", c_uint16),
        ("alpha_mask", c_uint16),
    )


class RenderPictforminfo(Structure):
    _fields_ = (
        ("id_", RenderPictformat),
        ("type_", c_uint8),
        ("depth", c_uint8),
        ("pad0", c_uint8 * 2),
        ("direct", RenderDirectformat),
        ("colormap", Colormap),
    )


class RenderPictvisual(Structure):
    _fields_ = (
        ("visual", Visualid),
        ("format_", RenderPictformat),
    )


class RenderPictdepth(Structure):
    _fields_ = (
        ("depth", c_uint8),
        ("pad0", c_uint8 * 1),
        ("num_visuals", c_uint16),
        ("pad1", c_uint8 * 4),
    )


class RenderPictdepthIterator(Structure):
    _fields_ = (("data", POINTER(RenderPictdepth)), ("rem", c_int), ("index", c_int))


class RenderPictscreen(Structure):
    _fields_ = (
        ("num_depths", c_uint32),
        ("fallback", RenderPictformat),
    )


class RenderPictscreenIterator(Structure):
    _fields_ = (("data", POINTER(RenderPictscreen)), ("rem", c_int), ("index", c_int))


class RenderQueryPictFormatsReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("pad0", c_uint8 * 1),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("num_formats", c_uint32),
        ("num_screens", c_uint32),
        ("num_depths", c_uint32),
        ("num_visuals", c_uint32),
        ("num_subpixel", c_uint32),
        ("pad1", c_uint8 * 4),
    )


class ShmQueryVersionReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("shared_pixmaps", c_uint8),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("major_version", c_uint16),
        ("minor_version", c_uint16),
        ("uid", c_uint16),
        ("gid", c_uint16),
        ("pixmap_format", c_uint8),
        ("pad0", c_uint8 * 15),
    )


class ShmSeg(XID):
    pass


class ShmGetImageReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("depth", c_uint8),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("visual", Visualid),
        ("size", c_uint32),
        ("pad0", c_uint8 * 16),
    )


class ShmCreateSegmentReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("nfd", c_uint8),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("pad0", c_uint8 * 24),
    )


class XfixesQueryVersionReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("pad0", c_uint8 * 1),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("major_version", c_uint32),
        ("minor_version", c_uint32),
        ("pad1", c_uint8 * 16),
    )


class XfixesGetCursorImageReply(Structure):
    _fields_ = (
        ("response_type", c_uint8),
        ("pad0", c_uint8 * 1),
        ("sequence", c_uint16),
        ("length", c_uint32),
        ("x", c_int16),
        ("y", c_int16),
        ("width", c_uint16),
        ("height", c_uint16),
        ("xhot", c_uint16),
        ("yhot", c_uint16),
        ("cursor_serial", c_uint32),
        ("pad1", c_uint8 * 8),
    )


def depth_visuals(r: Depth) -> Array[Visualtype]:
    return array_from_xcb(LIB.xcb.xcb_depth_visuals, LIB.xcb.xcb_depth_visuals_length, r)


def screen_allowed_depths(r: Screen) -> list[Depth]:
    return list_from_xcb(LIB.xcb.xcb_screen_allowed_depths_iterator, LIB.xcb.xcb_depth_next, r)


def setup_vendor(r: Setup) -> Array[c_char]:
    return array_from_xcb(LIB.xcb.xcb_setup_vendor, LIB.xcb.xcb_setup_vendor_length, r)


def setup_pixmap_formats(r: Setup) -> Array[Format]:
    return array_from_xcb(LIB.xcb.xcb_setup_pixmap_formats, LIB.xcb.xcb_setup_pixmap_formats_length, r)


def setup_roots(r: Setup) -> list[Screen]:
    return list_from_xcb(LIB.xcb.xcb_setup_roots_iterator, LIB.xcb.xcb_screen_next, r)


def get_image_data(r: GetImageReply) -> Array[c_uint8]:
    return array_from_xcb(LIB.xcb.xcb_get_image_data, LIB.xcb.xcb_get_image_data_length, r)


def get_property_value(r: GetPropertyReply) -> Array[c_char]:
    return array_from_xcb(LIB.xcb.xcb_get_property_value, LIB.xcb.xcb_get_property_value_length, r)


def randr_get_screen_resources_crtcs(r: RandrGetScreenResourcesReply) -> Array[RandrCrtc]:
    return array_from_xcb(
        LIB.randr.xcb_randr_get_screen_resources_crtcs, LIB.randr.xcb_randr_get_screen_resources_crtcs_length, r
    )


def randr_get_screen_resources_outputs(r: RandrGetScreenResourcesReply) -> Array[RandrOutput]:
    return array_from_xcb(
        LIB.randr.xcb_randr_get_screen_resources_outputs, LIB.randr.xcb_randr_get_screen_resources_outputs_length, r
    )


def randr_get_screen_resources_modes(r: RandrGetScreenResourcesReply) -> Array[RandrModeInfo]:
    return array_from_xcb(
        LIB.randr.xcb_randr_get_screen_resources_modes, LIB.randr.xcb_randr_get_screen_resources_modes_length, r
    )


def randr_get_screen_resources_names(r: RandrGetScreenResourcesReply) -> Array[c_uint8]:
    return array_from_xcb(
        LIB.randr.xcb_randr_get_screen_resources_names, LIB.randr.xcb_randr_get_screen_resources_names_length, r
    )


def randr_get_screen_resources_current_crtcs(r: RandrGetScreenResourcesCurrentReply) -> Array[RandrCrtc]:
    return array_from_xcb(
        LIB.randr.xcb_randr_get_screen_resources_current_crtcs,
        LIB.randr.xcb_randr_get_screen_resources_current_crtcs_length,
        r,
    )


def randr_get_screen_resources_current_outputs(r: RandrGetScreenResourcesCurrentReply) -> Array[RandrOutput]:
    return array_from_xcb(
        LIB.randr.xcb_randr_get_screen_resources_current_outputs,
        LIB.randr.xcb_randr_get_screen_resources_current_outputs_length,
        r,
    )


def randr_get_screen_resources_current_modes(r: RandrGetScreenResourcesCurrentReply) -> Array[RandrModeInfo]:
    return array_from_xcb(
        LIB.randr.xcb_randr_get_screen_resources_current_modes,
        LIB.randr.xcb_randr_get_screen_resources_current_modes_length,
        r,
    )


def randr_get_screen_resources_current_names(r: RandrGetScreenResourcesCurrentReply) -> Array[c_uint8]:
    return array_from_xcb(
        LIB.randr.xcb_randr_get_screen_resources_current_names,
        LIB.randr.xcb_randr_get_screen_resources_current_names_length,
        r,
    )


def randr_get_crtc_info_outputs(r: RandrGetCrtcInfoReply) -> Array[RandrOutput]:
    return array_from_xcb(
        LIB.randr.xcb_randr_get_crtc_info_outputs, LIB.randr.xcb_randr_get_crtc_info_outputs_length, r
    )


def randr_get_crtc_info_possible(r: RandrGetCrtcInfoReply) -> Array[RandrOutput]:
    return array_from_xcb(
        LIB.randr.xcb_randr_get_crtc_info_possible, LIB.randr.xcb_randr_get_crtc_info_possible_length, r
    )


def randr_get_output_info_crtcs(r: RandrGetOutputInfoReply) -> Array[RandrCrtc]:
    return array_from_xcb(
        LIB.randr.xcb_randr_get_output_info_crtcs, LIB.randr.xcb_randr_get_output_info_crtcs_length, r
    )


def randr_get_output_info_modes(r: RandrGetOutputInfoReply) -> Array[RandrMode]:
    return array_from_xcb(
        LIB.randr.xcb_randr_get_output_info_modes, LIB.randr.xcb_randr_get_output_info_modes_length, r
    )


def randr_get_output_info_clones(r: RandrGetOutputInfoReply) -> Array[RandrOutput]:
    return array_from_xcb(
        LIB.randr.xcb_randr_get_output_info_clones, LIB.randr.xcb_randr_get_output_info_clones_length, r
    )


def randr_get_output_info_name(r: RandrGetOutputInfoReply) -> Array[c_uint8]:
    return array_from_xcb(LIB.randr.xcb_randr_get_output_info_name, LIB.randr.xcb_randr_get_output_info_name_length, r)


def randr_get_output_property_data(r: RandrGetOutputPropertyReply) -> Array[c_uint8]:
    return array_from_xcb(
        LIB.randr.xcb_randr_get_output_property_data, LIB.randr.xcb_randr_get_output_property_data_length, r
    )


def randr_monitor_info_outputs(r: RandrMonitorInfo) -> Array[RandrOutput]:
    return array_from_xcb(LIB.randr.xcb_randr_monitor_info_outputs, LIB.randr.xcb_randr_monitor_info_outputs_length, r)


def randr_get_monitors_monitors(r: RandrGetMonitorsReply) -> list[RandrMonitorInfo]:
    return list_from_xcb(LIB.randr.xcb_randr_get_monitors_monitors_iterator, LIB.randr.xcb_randr_monitor_info_next, r)


def render_pictdepth_visuals(r: RenderPictdepth) -> Array[RenderPictvisual]:
    return array_from_xcb(LIB.render.xcb_render_pictdepth_visuals, LIB.render.xcb_render_pictdepth_visuals_length, r)


def render_pictscreen_depths(r: RenderPictscreen) -> list[RenderPictdepth]:
    return list_from_xcb(LIB.render.xcb_render_pictscreen_depths_iterator, LIB.render.xcb_render_pictdepth_next, r)


def render_query_pict_formats_formats(r: RenderQueryPictFormatsReply) -> Array[RenderPictforminfo]:
    return array_from_xcb(
        LIB.render.xcb_render_query_pict_formats_formats, LIB.render.xcb_render_query_pict_formats_formats_length, r
    )


def render_query_pict_formats_screens(r: RenderQueryPictFormatsReply) -> list[RenderPictscreen]:
    return list_from_xcb(
        LIB.render.xcb_render_query_pict_formats_screens_iterator, LIB.render.xcb_render_pictscreen_next, r
    )


def render_query_pict_formats_subpixels(r: RenderQueryPictFormatsReply) -> Array[c_uint32]:
    return array_from_xcb(
        LIB.render.xcb_render_query_pict_formats_subpixels, LIB.render.xcb_render_query_pict_formats_subpixels_length, r
    )


def xfixes_get_cursor_image_cursor_image(r: XfixesGetCursorImageReply) -> Array[c_uint32]:
    return array_from_xcb(
        LIB.xfixes.xcb_xfixes_get_cursor_image_cursor_image,
        LIB.xfixes.xcb_xfixes_get_cursor_image_cursor_image_length,
        r,
    )


def shm_create_segment_reply_fds(c: Connection | _Pointer[Connection], r: ShmCreateSegmentReply) -> _Pointer[c_int]:
    return LIB.shm.xcb_shm_create_segment_reply_fds(c, r)


def get_geometry(c: Connection, drawable: Drawable) -> GetGeometryReply:
    return LIB.xcb.xcb_get_geometry(c, drawable).reply(c)


def get_image(
    c: Connection,
    format_: c_uint8 | int,
    drawable: Drawable,
    x: c_int16 | int,
    y: c_int16 | int,
    width: c_uint16 | int,
    height: c_uint16 | int,
    plane_mask: c_uint32 | int,
) -> GetImageReply:
    return LIB.xcb.xcb_get_image(c, format_, drawable, x, y, width, height, plane_mask).reply(c)


def get_property(
    c: Connection,
    delete: c_uint8 | int,
    window: Window,
    property_: Atom,
    type_: Atom,
    long_offset: c_uint32 | int,
    long_length: c_uint32 | int,
) -> GetPropertyReply:
    return LIB.xcb.xcb_get_property(c, delete, window, property_, type_, long_offset, long_length).reply(c)


def no_operation(c: Connection) -> None:
    return LIB.xcb.xcb_no_operation_checked(c).check(c)


def randr_query_version(
    c: Connection, major_version: c_uint32 | int, minor_version: c_uint32 | int
) -> RandrQueryVersionReply:
    return LIB.randr.xcb_randr_query_version(c, major_version, minor_version).reply(c)


def randr_get_screen_resources(c: Connection, window: Window) -> RandrGetScreenResourcesReply:
    return LIB.randr.xcb_randr_get_screen_resources(c, window).reply(c)


def randr_get_screen_resources_current(c: Connection, window: Window) -> RandrGetScreenResourcesCurrentReply:
    return LIB.randr.xcb_randr_get_screen_resources_current(c, window).reply(c)


def randr_get_crtc_info(c: Connection, crtc: RandrCrtc, config_timestamp: Timestamp) -> RandrGetCrtcInfoReply:
    return LIB.randr.xcb_randr_get_crtc_info(c, crtc, config_timestamp).reply(c)


def randr_get_output_info(c: Connection, output: RandrOutput, config_timestamp: Timestamp) -> RandrGetOutputInfoReply:
    return LIB.randr.xcb_randr_get_output_info(c, output, config_timestamp).reply(c)


def randr_get_output_primary(c: Connection, window: Window) -> RandrGetOutputPrimaryReply:
    return LIB.randr.xcb_randr_get_output_primary(c, window).reply(c)


def randr_get_output_property(
    c: Connection,
    output: RandrOutput,
    property_: Atom,
    type_: Atom,
    long_offset: c_uint32 | int,
    long_length: c_uint32 | int,
    delete: c_uint8 | int,
    pending: c_uint8 | int,
) -> RandrGetOutputPropertyReply:
    return LIB.randr.xcb_randr_get_output_property(
        c, output, property_, type_, long_offset, long_length, delete, pending
    ).reply(c)


def randr_get_monitors(c: Connection, window: Window, get_active: c_uint8 | int) -> RandrGetMonitorsReply:
    return LIB.randr.xcb_randr_get_monitors(c, window, get_active).reply(c)


def render_query_version(
    c: Connection, client_major_version: c_uint32 | int, client_minor_version: c_uint32 | int
) -> RenderQueryVersionReply:
    return LIB.render.xcb_render_query_version(c, client_major_version, client_minor_version).reply(c)


def render_query_pict_formats(c: Connection) -> RenderQueryPictFormatsReply:
    return LIB.render.xcb_render_query_pict_formats(c).reply(c)


def shm_query_version(c: Connection) -> ShmQueryVersionReply:
    return LIB.shm.xcb_shm_query_version(c).reply(c)


def shm_get_image(
    c: Connection,
    drawable: Drawable,
    x: c_int16 | int,
    y: c_int16 | int,
    width: c_uint16 | int,
    height: c_uint16 | int,
    plane_mask: c_uint32 | int,
    format_: c_uint8 | int,
    shmseg: ShmSeg,
    offset: c_uint32 | int,
) -> ShmGetImageReply:
    return LIB.shm.xcb_shm_get_image(c, drawable, x, y, width, height, plane_mask, format_, shmseg, offset).reply(c)


def shm_attach_fd(c: Connection, shmseg: ShmSeg, shm_fd: c_int | int, read_only: c_uint8 | int) -> None:
    return LIB.shm.xcb_shm_attach_fd_checked(c, shmseg, shm_fd, read_only).check(c)


def shm_create_segment(
    c: Connection, shmseg: ShmSeg, size: c_uint32 | int, read_only: c_uint8 | int
) -> ShmCreateSegmentReply:
    return LIB.shm.xcb_shm_create_segment(c, shmseg, size, read_only).reply(c)


def shm_detach(c: Connection, shmseg: ShmSeg) -> None:
    return LIB.shm.xcb_shm_detach_checked(c, shmseg).check(c)


def xfixes_query_version(
    c: Connection, client_major_version: c_uint32 | int, client_minor_version: c_uint32 | int
) -> XfixesQueryVersionReply:
    return LIB.xfixes.xcb_xfixes_query_version(c, client_major_version, client_minor_version).reply(c)


def xfixes_get_cursor_image(c: Connection) -> XfixesGetCursorImageReply:
    return LIB.xfixes.xcb_xfixes_get_cursor_image(c).reply(c)


def initialize() -> None:  # noqa: PLR0915
    LIB.xcb.xcb_depth_next.argtypes = (POINTER(DepthIterator),)
    LIB.xcb.xcb_depth_next.restype = None
    LIB.xcb.xcb_screen_next.argtypes = (POINTER(ScreenIterator),)
    LIB.xcb.xcb_screen_next.restype = None
    LIB.xcb.xcb_setup_next.argtypes = (POINTER(SetupIterator),)
    LIB.xcb.xcb_setup_next.restype = None
    LIB.randr.xcb_randr_monitor_info_next.argtypes = (POINTER(RandrMonitorInfoIterator),)
    LIB.randr.xcb_randr_monitor_info_next.restype = None
    LIB.render.xcb_render_pictdepth_next.argtypes = (POINTER(RenderPictdepthIterator),)
    LIB.render.xcb_render_pictdepth_next.restype = None
    LIB.render.xcb_render_pictscreen_next.argtypes = (POINTER(RenderPictscreenIterator),)
    LIB.render.xcb_render_pictscreen_next.restype = None
    LIB.xcb.xcb_depth_visuals.argtypes = (POINTER(Depth),)
    LIB.xcb.xcb_depth_visuals.restype = POINTER(Visualtype)
    LIB.xcb.xcb_depth_visuals_length.argtypes = (POINTER(Depth),)
    LIB.xcb.xcb_depth_visuals_length.restype = c_int
    LIB.xcb.xcb_screen_allowed_depths_iterator.argtypes = (POINTER(Screen),)
    LIB.xcb.xcb_screen_allowed_depths_iterator.restype = DepthIterator
    LIB.xcb.xcb_setup_vendor.argtypes = (POINTER(Setup),)
    LIB.xcb.xcb_setup_vendor.restype = POINTER(c_char)
    LIB.xcb.xcb_setup_vendor_length.argtypes = (POINTER(Setup),)
    LIB.xcb.xcb_setup_vendor_length.restype = c_int
    LIB.xcb.xcb_setup_pixmap_formats.argtypes = (POINTER(Setup),)
    LIB.xcb.xcb_setup_pixmap_formats.restype = POINTER(Format)
    LIB.xcb.xcb_setup_pixmap_formats_length.argtypes = (POINTER(Setup),)
    LIB.xcb.xcb_setup_pixmap_formats_length.restype = c_int
    LIB.xcb.xcb_setup_roots_iterator.argtypes = (POINTER(Setup),)
    LIB.xcb.xcb_setup_roots_iterator.restype = ScreenIterator
    LIB.xcb.xcb_get_image_data.argtypes = (POINTER(GetImageReply),)
    LIB.xcb.xcb_get_image_data.restype = POINTER(c_uint8)
    LIB.xcb.xcb_get_image_data_length.argtypes = (POINTER(GetImageReply),)
    LIB.xcb.xcb_get_image_data_length.restype = c_int
    LIB.xcb.xcb_get_property_value.argtypes = (POINTER(GetPropertyReply),)
    LIB.xcb.xcb_get_property_value.restype = POINTER(c_char)
    LIB.xcb.xcb_get_property_value_length.argtypes = (POINTER(GetPropertyReply),)
    LIB.xcb.xcb_get_property_value_length.restype = c_int
    LIB.randr.xcb_randr_get_screen_resources_crtcs.argtypes = (POINTER(RandrGetScreenResourcesReply),)
    LIB.randr.xcb_randr_get_screen_resources_crtcs.restype = POINTER(RandrCrtc)
    LIB.randr.xcb_randr_get_screen_resources_crtcs_length.argtypes = (POINTER(RandrGetScreenResourcesReply),)
    LIB.randr.xcb_randr_get_screen_resources_crtcs_length.restype = c_int
    LIB.randr.xcb_randr_get_screen_resources_outputs.argtypes = (POINTER(RandrGetScreenResourcesReply),)
    LIB.randr.xcb_randr_get_screen_resources_outputs.restype = POINTER(RandrOutput)
    LIB.randr.xcb_randr_get_screen_resources_outputs_length.argtypes = (POINTER(RandrGetScreenResourcesReply),)
    LIB.randr.xcb_randr_get_screen_resources_outputs_length.restype = c_int
    LIB.randr.xcb_randr_get_screen_resources_modes.argtypes = (POINTER(RandrGetScreenResourcesReply),)
    LIB.randr.xcb_randr_get_screen_resources_modes.restype = POINTER(RandrModeInfo)
    LIB.randr.xcb_randr_get_screen_resources_modes_length.argtypes = (POINTER(RandrGetScreenResourcesReply),)
    LIB.randr.xcb_randr_get_screen_resources_modes_length.restype = c_int
    LIB.randr.xcb_randr_get_screen_resources_names.argtypes = (POINTER(RandrGetScreenResourcesReply),)
    LIB.randr.xcb_randr_get_screen_resources_names.restype = POINTER(c_uint8)
    LIB.randr.xcb_randr_get_screen_resources_names_length.argtypes = (POINTER(RandrGetScreenResourcesReply),)
    LIB.randr.xcb_randr_get_screen_resources_names_length.restype = c_int
    LIB.randr.xcb_randr_get_screen_resources_current_crtcs.argtypes = (POINTER(RandrGetScreenResourcesCurrentReply),)
    LIB.randr.xcb_randr_get_screen_resources_current_crtcs.restype = POINTER(RandrCrtc)
    LIB.randr.xcb_randr_get_screen_resources_current_crtcs_length.argtypes = (
        POINTER(RandrGetScreenResourcesCurrentReply),
    )
    LIB.randr.xcb_randr_get_screen_resources_current_crtcs_length.restype = c_int
    LIB.randr.xcb_randr_get_screen_resources_current_outputs.argtypes = (POINTER(RandrGetScreenResourcesCurrentReply),)
    LIB.randr.xcb_randr_get_screen_resources_current_outputs.restype = POINTER(RandrOutput)
    LIB.randr.xcb_randr_get_screen_resources_current_outputs_length.argtypes = (
        POINTER(RandrGetScreenResourcesCurrentReply),
    )
    LIB.randr.xcb_randr_get_screen_resources_current_outputs_length.restype = c_int
    LIB.randr.xcb_randr_get_screen_resources_current_modes.argtypes = (POINTER(RandrGetScreenResourcesCurrentReply),)
    LIB.randr.xcb_randr_get_screen_resources_current_modes.restype = POINTER(RandrModeInfo)
    LIB.randr.xcb_randr_get_screen_resources_current_modes_length.argtypes = (
        POINTER(RandrGetScreenResourcesCurrentReply),
    )
    LIB.randr.xcb_randr_get_screen_resources_current_modes_length.restype = c_int
    LIB.randr.xcb_randr_get_screen_resources_current_names.argtypes = (POINTER(RandrGetScreenResourcesCurrentReply),)
    LIB.randr.xcb_randr_get_screen_resources_current_names.restype = POINTER(c_uint8)
    LIB.randr.xcb_randr_get_screen_resources_current_names_length.argtypes = (
        POINTER(RandrGetScreenResourcesCurrentReply),
    )
    LIB.randr.xcb_randr_get_screen_resources_current_names_length.restype = c_int
    LIB.randr.xcb_randr_get_crtc_info_outputs.argtypes = (POINTER(RandrGetCrtcInfoReply),)
    LIB.randr.xcb_randr_get_crtc_info_outputs.restype = POINTER(RandrOutput)
    LIB.randr.xcb_randr_get_crtc_info_outputs_length.argtypes = (POINTER(RandrGetCrtcInfoReply),)
    LIB.randr.xcb_randr_get_crtc_info_outputs_length.restype = c_int
    LIB.randr.xcb_randr_get_crtc_info_possible.argtypes = (POINTER(RandrGetCrtcInfoReply),)
    LIB.randr.xcb_randr_get_crtc_info_possible.restype = POINTER(RandrOutput)
    LIB.randr.xcb_randr_get_crtc_info_possible_length.argtypes = (POINTER(RandrGetCrtcInfoReply),)
    LIB.randr.xcb_randr_get_crtc_info_possible_length.restype = c_int
    LIB.randr.xcb_randr_get_output_info_crtcs.argtypes = (POINTER(RandrGetOutputInfoReply),)
    LIB.randr.xcb_randr_get_output_info_crtcs.restype = POINTER(RandrCrtc)
    LIB.randr.xcb_randr_get_output_info_crtcs_length.argtypes = (POINTER(RandrGetOutputInfoReply),)
    LIB.randr.xcb_randr_get_output_info_crtcs_length.restype = c_int
    LIB.randr.xcb_randr_get_output_info_modes.argtypes = (POINTER(RandrGetOutputInfoReply),)
    LIB.randr.xcb_randr_get_output_info_modes.restype = POINTER(RandrMode)
    LIB.randr.xcb_randr_get_output_info_modes_length.argtypes = (POINTER(RandrGetOutputInfoReply),)
    LIB.randr.xcb_randr_get_output_info_modes_length.restype = c_int
    LIB.randr.xcb_randr_get_output_info_clones.argtypes = (POINTER(RandrGetOutputInfoReply),)
    LIB.randr.xcb_randr_get_output_info_clones.restype = POINTER(RandrOutput)
    LIB.randr.xcb_randr_get_output_info_clones_length.argtypes = (POINTER(RandrGetOutputInfoReply),)
    LIB.randr.xcb_randr_get_output_info_clones_length.restype = c_int
    LIB.randr.xcb_randr_get_output_info_name.argtypes = (POINTER(RandrGetOutputInfoReply),)
    LIB.randr.xcb_randr_get_output_info_name.restype = POINTER(c_uint8)
    LIB.randr.xcb_randr_get_output_info_name_length.argtypes = (POINTER(RandrGetOutputInfoReply),)
    LIB.randr.xcb_randr_get_output_info_name_length.restype = c_int
    LIB.randr.xcb_randr_get_output_property_data.argtypes = (POINTER(RandrGetOutputPropertyReply),)
    LIB.randr.xcb_randr_get_output_property_data.restype = POINTER(c_uint8)
    LIB.randr.xcb_randr_get_output_property_data_length.argtypes = (POINTER(RandrGetOutputPropertyReply),)
    LIB.randr.xcb_randr_get_output_property_data_length.restype = c_int
    LIB.randr.xcb_randr_monitor_info_outputs.argtypes = (POINTER(RandrMonitorInfo),)
    LIB.randr.xcb_randr_monitor_info_outputs.restype = POINTER(RandrOutput)
    LIB.randr.xcb_randr_monitor_info_outputs_length.argtypes = (POINTER(RandrMonitorInfo),)
    LIB.randr.xcb_randr_monitor_info_outputs_length.restype = c_int
    LIB.randr.xcb_randr_get_monitors_monitors_iterator.argtypes = (POINTER(RandrGetMonitorsReply),)
    LIB.randr.xcb_randr_get_monitors_monitors_iterator.restype = RandrMonitorInfoIterator
    LIB.render.xcb_render_pictdepth_visuals.argtypes = (POINTER(RenderPictdepth),)
    LIB.render.xcb_render_pictdepth_visuals.restype = POINTER(RenderPictvisual)
    LIB.render.xcb_render_pictdepth_visuals_length.argtypes = (POINTER(RenderPictdepth),)
    LIB.render.xcb_render_pictdepth_visuals_length.restype = c_int
    LIB.render.xcb_render_pictscreen_depths_iterator.argtypes = (POINTER(RenderPictscreen),)
    LIB.render.xcb_render_pictscreen_depths_iterator.restype = RenderPictdepthIterator
    LIB.render.xcb_render_query_pict_formats_formats.argtypes = (POINTER(RenderQueryPictFormatsReply),)
    LIB.render.xcb_render_query_pict_formats_formats.restype = POINTER(RenderPictforminfo)
    LIB.render.xcb_render_query_pict_formats_formats_length.argtypes = (POINTER(RenderQueryPictFormatsReply),)
    LIB.render.xcb_render_query_pict_formats_formats_length.restype = c_int
    LIB.render.xcb_render_query_pict_formats_screens_iterator.argtypes = (POINTER(RenderQueryPictFormatsReply),)
    LIB.render.xcb_render_query_pict_formats_screens_iterator.restype = RenderPictscreenIterator
    LIB.render.xcb_render_query_pict_formats_subpixels.argtypes = (POINTER(RenderQueryPictFormatsReply),)
    LIB.render.xcb_render_query_pict_formats_subpixels.restype = POINTER(c_uint32)
    LIB.render.xcb_render_query_pict_formats_subpixels_length.argtypes = (POINTER(RenderQueryPictFormatsReply),)
    LIB.render.xcb_render_query_pict_formats_subpixels_length.restype = c_int
    LIB.xfixes.xcb_xfixes_get_cursor_image_cursor_image.argtypes = (POINTER(XfixesGetCursorImageReply),)
    LIB.xfixes.xcb_xfixes_get_cursor_image_cursor_image.restype = POINTER(c_uint32)
    LIB.xfixes.xcb_xfixes_get_cursor_image_cursor_image_length.argtypes = (POINTER(XfixesGetCursorImageReply),)
    LIB.xfixes.xcb_xfixes_get_cursor_image_cursor_image_length.restype = c_int
    LIB.shm.xcb_shm_create_segment_reply_fds.argtypes = (POINTER(Connection), POINTER(ShmCreateSegmentReply))
    LIB.shm.xcb_shm_create_segment_reply_fds.restype = POINTER(c_int)
    initialize_xcb_typed_func(LIB.xcb, "xcb_get_geometry", [POINTER(Connection), Drawable], GetGeometryReply)
    initialize_xcb_typed_func(
        LIB.xcb,
        "xcb_get_image",
        [POINTER(Connection), c_uint8, Drawable, c_int16, c_int16, c_uint16, c_uint16, c_uint32],
        GetImageReply,
    )
    initialize_xcb_typed_func(
        LIB.xcb,
        "xcb_get_property",
        [POINTER(Connection), c_uint8, Window, Atom, Atom, c_uint32, c_uint32],
        GetPropertyReply,
    )
    LIB.xcb.xcb_no_operation_checked.argtypes = (POINTER(Connection),)
    LIB.xcb.xcb_no_operation_checked.restype = VoidCookie
    initialize_xcb_typed_func(
        LIB.randr, "xcb_randr_query_version", [POINTER(Connection), c_uint32, c_uint32], RandrQueryVersionReply
    )
    initialize_xcb_typed_func(
        LIB.randr, "xcb_randr_get_screen_resources", [POINTER(Connection), Window], RandrGetScreenResourcesReply
    )
    initialize_xcb_typed_func(
        LIB.randr,
        "xcb_randr_get_screen_resources_current",
        [POINTER(Connection), Window],
        RandrGetScreenResourcesCurrentReply,
    )
    initialize_xcb_typed_func(
        LIB.randr, "xcb_randr_get_crtc_info", [POINTER(Connection), RandrCrtc, Timestamp], RandrGetCrtcInfoReply
    )
    initialize_xcb_typed_func(
        LIB.randr, "xcb_randr_get_output_info", [POINTER(Connection), RandrOutput, Timestamp], RandrGetOutputInfoReply
    )
    initialize_xcb_typed_func(
        LIB.randr, "xcb_randr_get_output_primary", [POINTER(Connection), Window], RandrGetOutputPrimaryReply
    )
    initialize_xcb_typed_func(
        LIB.randr,
        "xcb_randr_get_output_property",
        [POINTER(Connection), RandrOutput, Atom, Atom, c_uint32, c_uint32, c_uint8, c_uint8],
        RandrGetOutputPropertyReply,
    )
    initialize_xcb_typed_func(
        LIB.randr, "xcb_randr_get_monitors", [POINTER(Connection), Window, c_uint8], RandrGetMonitorsReply
    )
    initialize_xcb_typed_func(
        LIB.render, "xcb_render_query_version", [POINTER(Connection), c_uint32, c_uint32], RenderQueryVersionReply
    )
    initialize_xcb_typed_func(
        LIB.render, "xcb_render_query_pict_formats", [POINTER(Connection)], RenderQueryPictFormatsReply
    )
    initialize_xcb_typed_func(LIB.shm, "xcb_shm_query_version", [POINTER(Connection)], ShmQueryVersionReply)
    initialize_xcb_typed_func(
        LIB.shm,
        "xcb_shm_get_image",
        [POINTER(Connection), Drawable, c_int16, c_int16, c_uint16, c_uint16, c_uint32, c_uint8, ShmSeg, c_uint32],
        ShmGetImageReply,
    )
    LIB.shm.xcb_shm_attach_fd_checked.argtypes = (POINTER(Connection), ShmSeg, c_int, c_uint8)
    LIB.shm.xcb_shm_attach_fd_checked.restype = VoidCookie
    initialize_xcb_typed_func(
        LIB.shm, "xcb_shm_create_segment", [POINTER(Connection), ShmSeg, c_uint32, c_uint8], ShmCreateSegmentReply
    )
    LIB.shm.xcb_shm_detach_checked.argtypes = (POINTER(Connection), ShmSeg)
    LIB.shm.xcb_shm_detach_checked.restype = VoidCookie
    initialize_xcb_typed_func(
        LIB.xfixes, "xcb_xfixes_query_version", [POINTER(Connection), c_uint32, c_uint32], XfixesQueryVersionReply
    )
    initialize_xcb_typed_func(
        LIB.xfixes, "xcb_xfixes_get_cursor_image", [POINTER(Connection)], XfixesGetCursorImageReply
    )



# This is part of the MSS Python's module.
# Source: https://github.com/BoboTiG/python-mss.

from __future__ import annotations

from typing import TYPE_CHECKING

from mss.exception import ScreenShotError
from mss.models import Monitor, Pixel, Pixels, Pos, Size

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Any


class ScreenShot:
    """Screenshot object.

    .. note::

        A better name would have  been *Image*, but to prevent collisions
        with PIL.Image, it has been decided to use *ScreenShot*.
    """

    __slots__ = {"__pixels", "__rgb", "pos", "raw", "size"}

    def __init__(self, data: bytearray, monitor: Monitor, /, *, size: Size | None = None) -> None:
        self.__pixels: Pixels | None = None
        self.__rgb: bytes | None = None

        #: Bytearray of the raw BGRA pixels retrieved by ctypes
        #: OS independent implementations.
        self.raw: bytearray = data

        #: NamedTuple of the screenshot coordinates.
        self.pos: Pos = Pos(monitor["left"], monitor["top"])

        #: NamedTuple of the screenshot size.
        self.size: Size = Size(monitor["width"], monitor["height"]) if size is None else size

    def __repr__(self) -> str:
        return f"<{type(self).__name__} pos={self.left},{self.top} size={self.width}x{self.height}>"

    @property
    def __array_interface__(self) -> dict[str, Any]:
        """NumPy array interface support.

        This is used by NumPy, many SciPy projects, CuPy, PyTorch (via
        ``torch.from_numpy``), TensorFlow (via ``tf.convert_to_tensor``),
        JAX (via ``jax.numpy.asarray``), Pandas, scikit-learn, Matplotlib,
        some OpenCV functions, and others.  This allows you to pass a
        :class:`ScreenShot` instance directly to these libraries without
        needing to convert it first.

        This is in HWC order, with 4 channels (BGRA).

        .. seealso::

            https://numpy.org/doc/stable/reference/arrays.interface.html
               The NumPy array interface protocol specification
        """
        return {
            "version": 3,
            "shape": (self.height, self.width, 4),
            "typestr": "|u1",
            "data": self.raw,
        }

    @classmethod
    def from_size(cls: type[ScreenShot], data: bytearray, width: int, height: int, /) -> ScreenShot:
        """Instantiate a new class given only screenshot's data and size."""
        monitor = {"left": 0, "top": 0, "width": width, "height": height}
        return cls(data, monitor)

    @property
    def bgra(self) -> bytes:
        """BGRx values from the BGRx raw pixels.

        The format is a bytes object with BGRxBGRx... sequence.  A specific
        pixel can be accessed as
        ``bgra[(y * width + x) * 4:(y * width + x) * 4 + 4].``

        .. note::
            While the name is ``bgra``, the alpha channel may or may not be
            valid.
        """
        return bytes(self.raw)

    @property
    def pixels(self) -> Pixels:
        """RGB tuples.

        The format is a list of rows.  Each row is a list of pixels.
        Each pixel is a tuple of (R, G, B).
        """
        if not self.__pixels:
            rgb_tuples: Iterator[Pixel] = zip(self.raw[2::4], self.raw[1::4], self.raw[::4])
            self.__pixels = list(zip(*[iter(rgb_tuples)] * self.width))

        return self.__pixels

    def pixel(self, coord_x: int, coord_y: int) -> Pixel:
        """Return the pixel value at a given position.

        :returns: A tuple of (R, G, B) values.
        """
        try:
            return self.pixels[coord_y][coord_x]
        except IndexError as exc:
            msg = f"Pixel location ({coord_x}, {coord_y}) is out of range."
            raise ScreenShotError(msg) from exc

    @property
    def rgb(self) -> bytes:
        """Compute RGB values from the BGRA raw pixels.

        The format is a bytes object with BGRBGR... sequence.  A specific
        pixel can be accessed as
        ``rgb[(y * width + x) * 3:(y * width + x) * 3 + 3]``.
        """
        if not self.__rgb:
            rgb = bytearray(self.height * self.width * 3)
            raw = self.raw
            rgb[::3] = raw[2::4]
            rgb[1::3] = raw[1::4]
            rgb[2::3] = raw[::4]
            self.__rgb = bytes(rgb)

        return self.__rgb

    @property
    def top(self) -> int:
        """Convenient accessor to the top position."""
        return self.pos.top

    @property
    def left(self) -> int:
        """Convenient accessor to the left position."""
        return self.pos.left

    @property
    def width(self) -> int:
        """Convenient accessor to the width size."""
        return self.size.width

    @property
    def height(self) -> int:
        """Convenient accessor to the height size."""
        return self.size.height



# This is part of the MSS Python's module.
# Source: https://github.com/BoboTiG/python-mss.

from __future__ import annotations

from typing import Any


class ScreenShotError(Exception):
    """Error handling class."""

    def __init__(self, message: str, /, *, details: dict[str, Any] | None = None) -> None:
        super().__init__(message)
        #: On GNU/Linux, and if the error comes from the XServer, it contains XError details.
        #: This is an empty dict by default.
        #:
        #: For XErrors, you can find information on
        #: `Using the Default Error Handlers <https://tronche.com/gui/x/xlib/event-handling/protocol-errors/default-handlers.html>`_.
        #:
        #: .. versionadded:: 3.3.0
        self.details = details or {}



from __future__ import annotations

import contextlib
from ctypes import _Pointer, addressof, c_int
from typing import Literal, overload

from mss.linux import xcbgen

# We import these just so they're re-exported to our users.
# ruff: noqa: F401
from mss.linux.xcbgen import (
    RANDR_MAJOR_VERSION,
    RANDR_MINOR_VERSION,
    RENDER_MAJOR_VERSION,
    RENDER_MINOR_VERSION,
    SHM_MAJOR_VERSION,
    SHM_MINOR_VERSION,
    XFIXES_MAJOR_VERSION,
    XFIXES_MINOR_VERSION,
    Atom,
    BackingStore,
    Colormap,
    Depth,
    DepthIterator,
    Drawable,
    Format,
    GetGeometryReply,
    GetImageReply,
    GetPropertyReply,
    ImageFormat,
    ImageOrder,
    Keycode,
    Pixmap,
    RandrConnection,
    RandrCrtc,
    RandrGetCrtcInfoReply,
    RandrGetMonitorsReply,
    RandrGetOutputInfoReply,
    RandrGetOutputPrimaryReply,
    RandrGetOutputPropertyReply,
    RandrGetScreenResourcesCurrentReply,
    RandrGetScreenResourcesReply,
    RandrMode,
    RandrModeInfo,
    RandrMonitorInfo,
    RandrMonitorInfoIterator,
    RandrOutput,
    RandrQueryVersionReply,
    RandrSetConfig,
    RenderDirectformat,
    RenderPictdepth,
    RenderPictdepthIterator,
    RenderPictformat,
    RenderPictforminfo,
    RenderPictscreen,
    RenderPictscreenIterator,
    RenderPictType,
    RenderPictvisual,
    RenderQueryPictFormatsReply,
    RenderQueryVersionReply,
    RenderSubPixel,
    Screen,
    ScreenIterator,
    Setup,
    SetupIterator,
    ShmCreateSegmentReply,
    ShmGetImageReply,
    ShmQueryVersionReply,
    ShmSeg,
    Timestamp,
    VisualClass,
    Visualid,
    Visualtype,
    Window,
    XfixesGetCursorImageReply,
    XfixesQueryVersionReply,
    depth_visuals,
    get_geometry,
    get_image,
    get_image_data,
    get_property,
    get_property_value,
    no_operation,
    randr_get_crtc_info,
    randr_get_crtc_info_outputs,
    randr_get_crtc_info_possible,
    randr_get_monitors,
    randr_get_monitors_monitors,
    randr_get_output_info,
    randr_get_output_info_clones,
    randr_get_output_info_crtcs,
    randr_get_output_info_modes,
    randr_get_output_info_name,
    randr_get_output_primary,
    randr_get_output_property,
    randr_get_output_property_data,
    randr_get_screen_resources,
    randr_get_screen_resources_crtcs,
    randr_get_screen_resources_current,
    randr_get_screen_resources_current_crtcs,
    randr_get_screen_resources_current_modes,
    randr_get_screen_resources_current_names,
    randr_get_screen_resources_current_outputs,
    randr_get_screen_resources_modes,
    randr_get_screen_resources_names,
    randr_get_screen_resources_outputs,
    randr_monitor_info_outputs,
    randr_query_version,
    render_pictdepth_visuals,
    render_pictscreen_depths,
    render_query_pict_formats,
    render_query_pict_formats_formats,
    render_query_pict_formats_screens,
    render_query_pict_formats_subpixels,
    render_query_version,
    screen_allowed_depths,
    setup_pixmap_formats,
    setup_roots,
    setup_vendor,
    shm_attach_fd,
    shm_create_segment,
    shm_create_segment_reply_fds,
    shm_detach,
    shm_get_image,
    shm_query_version,
    xfixes_get_cursor_image,
    xfixes_get_cursor_image_cursor_image,
    xfixes_query_version,
)

# These are also here to re-export.
from mss.linux.xcbhelpers import LIB, XID, Connection, InternAtomReply, QueryExtensionReply, XcbExtension, XError

XCB_CONN_ERROR = 1
XCB_CONN_CLOSED_EXT_NOTSUPPORTED = 2
XCB_CONN_CLOSED_MEM_INSUFFICIENT = 3
XCB_CONN_CLOSED_REQ_LEN_EXCEED = 4
XCB_CONN_CLOSED_PARSE_ERR = 5
XCB_CONN_CLOSED_INVALID_SCREEN = 6
XCB_CONN_CLOSED_FDPASSING_FAILED = 7

# I don't know of error descriptions for the XCB connection errors being accessible through a library (a la strerror),
# and the ones in xcb.h's comments aren't too great, so I wrote these.
XCB_CONN_ERRMSG = {
    XCB_CONN_ERROR: "connection lost or could not be established",
    XCB_CONN_CLOSED_EXT_NOTSUPPORTED: "extension not supported",
    XCB_CONN_CLOSED_MEM_INSUFFICIENT: "memory exhausted",
    XCB_CONN_CLOSED_REQ_LEN_EXCEED: "request length longer than server accepts",
    XCB_CONN_CLOSED_PARSE_ERR: "display is unset or invalid (check $DISPLAY)",
    XCB_CONN_CLOSED_INVALID_SCREEN: "server does not have a screen matching the requested display",
    XCB_CONN_CLOSED_FDPASSING_FAILED: "could not pass file descriptor",
}


#### High-level XCB function wrappers

# XCB_NONE is the universal null resource or null atom parameter value for many core X requests
XCB_NONE = Atom(0)

# Some atoms are defined by the spec, to avoid apps having to look them up.  It's fine to look them up anyway.
_PREDEFINED_ATOMS = {
    "PRIMARY": Atom(1),
    "SECONDARY": Atom(2),
    "ARC": Atom(3),
    "ATOM": Atom(4),
    "BITMAP": Atom(5),
    "CARDINAL": Atom(6),
    "COLORMAP": Atom(7),
    "CURSOR": Atom(8),
    "CUT_BUFFER0": Atom(9),
    "CUT_BUFFER1": Atom(10),
    "CUT_BUFFER2": Atom(11),
    "CUT_BUFFER3": Atom(12),
    "CUT_BUFFER4": Atom(13),
    "CUT_BUFFER5": Atom(14),
    "CUT_BUFFER6": Atom(15),
    "CUT_BUFFER7": Atom(16),
    "DRAWABLE": Atom(17),
    "FONT": Atom(18),
    "INTEGER": Atom(19),
    "PIXMAP": Atom(20),
    "POINT": Atom(21),
    "RECTANGLE": Atom(22),
    "RESOURCE_MANAGER": Atom(23),
    "RGB_COLOR_MAP": Atom(24),
    "RGB_BEST_MAP": Atom(25),
    "RGB_BLUE_MAP": Atom(26),
    "RGB_DEFAULT_MAP": Atom(27),
    "RGB_GRAY_MAP": Atom(28),
    "RGB_GREEN_MAP": Atom(29),
    "RGB_RED_MAP": Atom(30),
    "STRING": Atom(31),
    "VISUALID": Atom(32),
    "WINDOW": Atom(33),
    "WM_COMMAND": Atom(34),
    "WM_HINTS": Atom(35),
    "WM_CLIENT_MACHINE": Atom(36),
    "WM_ICON_NAME": Atom(37),
    "WM_ICON_SIZE": Atom(38),
    "WM_NAME": Atom(39),
    "WM_NORMAL_HINTS": Atom(40),
    "WM_SIZE_HINTS": Atom(41),
    "WM_ZOOM_HINTS": Atom(42),
    "MIN_SPACE": Atom(43),
    "NORM_SPACE": Atom(44),
    "MAX_SPACE": Atom(45),
    "END_SPACE": Atom(46),
    "SUPERSCRIPT_X": Atom(47),
    "SUPERSCRIPT_Y": Atom(48),
    "SUBSCRIPT_X": Atom(49),
    "SUBSCRIPT_Y": Atom(50),
    "UNDERLINE_POSITION": Atom(51),
    "UNDERLINE_THICKNESS": Atom(52),
    "STRIKEOUT_ASCENT": Atom(53),
    "STRIKEOUT_DESCENT": Atom(54),
    "ITALIC_ANGLE": Atom(55),
    "X_HEIGHT": Atom(56),
    "QUAD_WIDTH": Atom(57),
    "WEIGHT": Atom(58),
    "POINT_SIZE": Atom(59),
    "RESOLUTION": Atom(60),
    "COPYRIGHT": Atom(61),
    "NOTICE": Atom(62),
    "FONT_NAME": Atom(63),
    "FAMILY_NAME": Atom(64),
    "FULL_NAME": Atom(65),
    "CAP_HEIGHT": Atom(66),
    "WM_CLASS": Atom(67),
    "WM_TRANSIENT_FOR": Atom(68),
}

# The atom cache needs to be per-connection. Rather than keying on a (connection, name) tuple, we use a two-level
# cache keyed by the integer address of the underlying XCB connection (see ctypes.addressof in intern_atom).
_ATOM_CACHE: dict[int, dict[str, Atom]] = {}


@overload
def intern_atom(
    xcb_conn: Connection | _Pointer[Connection],
    name: str,
    *,
    only_if_exists: Literal[False] = False,
) -> Atom: ...
@overload
def intern_atom(
    xcb_conn: Connection | _Pointer[Connection],
    name: str,
    *,
    only_if_exists: Literal[True] = True,
) -> Atom | None: ...
@overload
def intern_atom(
    xcb_conn: Connection | _Pointer[Connection],
    name: str,
    *,
    only_if_exists: bool,
) -> Atom | None: ...


def intern_atom(
    xcb_conn: Connection | _Pointer[Connection],
    name: str,
    *,
    only_if_exists: bool = False,
) -> Atom | None:
    if name in _PREDEFINED_ATOMS:
        return _PREDEFINED_ATOMS[name]

    if isinstance(xcb_conn, _Pointer):
        # Dereference the pointer before using the cache.
        xcb_conn = xcb_conn.contents
    cache_key = addressof(xcb_conn)
    if cache_key not in _ATOM_CACHE:
        # This can happen if the connection was closed and its cache cleared, but some code still has a reference to
        # the connection object.  We could re-create the cache entry, but it's safer to just fail instead of silently
        # allowing lookups to succeed when they shouldn't.
        msg = "Connection to X server is closed"
        raise XError(msg)
    if name in _ATOM_CACHE[cache_key]:
        return _ATOM_CACHE[cache_key][name]

    # Atom names are required to be Latin-1, per the X protocol spec, although anything that's not in the XPCS (a
    # subset of ASCII) is vendor-defined.
    name_encoded = name.encode("latin_1", errors="strict")
    cookie = LIB.xcb.xcb_intern_atom(xcb_conn, 1 if only_if_exists else 0, len(name_encoded), name_encoded)
    atom_as_xid = cookie.reply(xcb_conn).atom
    if atom_as_xid.value == 0:
        if not only_if_exists:
            # This shouldn't be possible.  We at least need to have a path for the type-checker to be happy, though.
            msg = f"X server failed to intern atom '{name}'"
            raise XError(msg)
        # We don't do negative caching, since any app might intern the atom at any time.
        return None
    atom = Atom(atom_as_xid.value)
    _ATOM_CACHE[cache_key][name] = atom
    return atom


def get_extension_data(
    xcb_conn: Connection | _Pointer[Connection], ext: XcbExtension | _Pointer[XcbExtension]
) -> QueryExtensionReply:
    """Get extension data for the given extension.

    Returns the extension data, which includes whether the extension is present
    and its opcode information.
    """
    reply_p = LIB.xcb.xcb_get_extension_data(xcb_conn, ext)
    return reply_p.contents


def prefetch_extension_data(
    xcb_conn: Connection | _Pointer[Connection], ext: XcbExtension | _Pointer[XcbExtension]
) -> None:
    """Prefetch extension data for the given extension.

    This is a performance hint to XCB to fetch the extension data
    asynchronously.
    """
    LIB.xcb.xcb_prefetch_extension_data(xcb_conn, ext)


def generate_id(xcb_conn: Connection | _Pointer[Connection]) -> XID:
    """Generate a new unique X resource ID.

    Returns an XID that can be used to create new X resources.
    """
    return LIB.xcb.xcb_generate_id(xcb_conn)


def get_setup(xcb_conn: Connection | _Pointer[Connection]) -> Setup:
    """Get the connection setup information.

    Returns the setup structure containing information about the X server,
    including available screens, pixmap formats, etc.
    """
    setup_p = LIB.xcb.xcb_get_setup(xcb_conn)
    return setup_p.contents


# Connection management


def initialize() -> None:
    LIB.initialize(callbacks=[xcbgen.initialize])


def connect(display: str | bytes | None = None) -> tuple[Connection, int]:
    if isinstance(display, str):
        display = display.encode("utf-8")

    initialize()
    pref_screen_num = c_int()
    conn_p = LIB.xcb.xcb_connect(display, pref_screen_num)

    # We still get a connection object even if the connection fails.
    conn_err = LIB.xcb.xcb_connection_has_error(conn_p)
    if conn_err != 0:
        # XCB won't free its connection structures until we disconnect, even in the event of an error.
        LIB.xcb.xcb_disconnect(conn_p)
        msg = "Cannot connect to display: "
        conn_errmsg = XCB_CONN_ERRMSG.get(conn_err)
        msg += conn_errmsg or f"error code {conn_err}"
        raise XError(msg)

    # Prefetch extension data for all extensions we support to populate XCB's internal cache.
    prefetch_extension_data(conn_p, LIB.randr_id)
    prefetch_extension_data(conn_p, LIB.render_id)
    prefetch_extension_data(conn_p, LIB.shm_id)
    prefetch_extension_data(conn_p, LIB.xfixes_id)

    _ATOM_CACHE[addressof(conn_p.contents)] = {}

    return conn_p.contents, pref_screen_num.value


def disconnect(xcb_conn: Connection | _Pointer[Connection]) -> None:
    if isinstance(xcb_conn, _Pointer):
        # Dereference the pointer before using the cache.
        xcb_conn = xcb_conn.contents

    # The cache might already be cleared if the connection had an error, or if disconnect was called multiple times.
    with contextlib.suppress(KeyError):
        del _ATOM_CACHE[addressof(xcb_conn)]

    conn_err = LIB.xcb.xcb_connection_has_error(xcb_conn)
    # XCB won't free its connection structures until we disconnect, even in the event of an error.
    LIB.xcb.xcb_disconnect(xcb_conn)
    if conn_err != 0:
        msg = "Connection to X server closed: "
        conn_errmsg = XCB_CONN_ERRMSG.get(conn_err)
        msg += conn_errmsg or f"error code {conn_err}"
        raise XError(msg)



# This module is maintained by Mickaël Schoentgen <contact@tiger-222.fr>.
#
# You can always get the latest version of this module at:
#     https://github.com/BoboTiG/python-mss
# If that URL should fail, try contacting the author.
"""An ultra fast cross-platform multiple screenshots module in pure python
using ctypes.
"""

from mss.base import MSS, ScreenShot
from mss.exception import ScreenShotError
from mss.factory import mss

__version__ = "10.2.0"
__author__ = "Mickaël Schoentgen"
__date__ = "2013-2026"
__copyright__ = f"""
Copyright (c) {__date__}, {__author__}

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee or royalty is hereby
granted, provided that the above copyright notice appear in all copies
and that both that copyright notice and this permission notice appear
in supporting documentation or portions thereof, including
modifications, that you make.
"""
__all__ = ("MSS", "ScreenShot", "ScreenShotError", "mss")



"""XCB-based backend using the XGetImage request.

This backend issues XCB ``GetImage`` requests and supports the RandR and
XFixes extensions when available for monitor enumeration and cursor capture.

This backend will work on any X connection, but is slower than the xshmgetimage
backend.

.. versionadded:: 10.2.0
"""

from mss.linux.base import MSSImplXCBBase
from mss.models import Monitor

__all__ = ()


class MSSImplXGetImage(MSSImplXCBBase):
    """XCB backend using XGetImage requests on GNU/Linux.

    .. seealso::
        :py:class:`mss.linux.base.MSSXCBBase`
            Lists constructor parameters.
    """

    def grab(self, monitor: Monitor) -> bytearray:
        """Retrieve all pixels from a monitor. Pixels have to be RGBX."""
        return super()._grab_xgetimage(monitor)



from __future__ import annotations

from typing import TYPE_CHECKING, Any
from urllib.parse import urlencode

from mss.base import MSSImplementation
from mss.exception import ScreenShotError
from mss.linux import xcb
from mss.linux.xcb import LIB
from mss.screenshot import ScreenShot
from mss.tools import parse_edid

if TYPE_CHECKING:
    from ctypes import Array

    from mss.models import Monitor, Monitors

__all__ = ()

SUPPORTED_DEPTHS = {24, 32}
SUPPORTED_BITS_PER_PIXEL = 32
SUPPORTED_RED_MASK = 0xFF0000
SUPPORTED_GREEN_MASK = 0x00FF00
SUPPORTED_BLUE_MASK = 0x0000FF
ALL_PLANES = 0xFFFFFFFF  # XCB doesn't define AllPlanes


class MSSImplXCBBase(MSSImplementation):
    """Base class for XCB-based screenshot implementations.

    Provides common XCB initialization and monitor detection logic that can be
    shared across different XCB screenshot methods (``XGetImage``,
    ``XShmGetImage``, ``XComposite``, etc.).

    :param display: Optional keyword argument.
        Specifies an X11 display string to connect to.  The default is
        taken from the environment variable :envvar:`DISPLAY`.
    :type display: str | bytes | None

    .. seealso::
        :py:class:`mss.MSS`
            Lists other parameters.
    """

    def __init__(self, *, display: str | bytes | None = None, with_cursor: bool = False) -> None:  # noqa: PLR0912
        super().__init__(with_cursor=with_cursor)

        if not display:
            display = None
        elif isinstance(display, str):
            display = display.encode("utf-8")

        self.conn: xcb.Connection | None
        self.conn, pref_screen_num = xcb.connect(display)

        # Get the connection setup information that was included when we connected.
        xcb_setup = xcb.get_setup(self.conn)
        screens = xcb.setup_roots(xcb_setup)
        # pref_screen_num is the screen object corresponding to the screen number, e.g., 1 if DISPLAY=":0.1".  It's
        # almost always the only screen (screen 0); nobody uses separate screens (in the X sense) anymore.
        self.pref_screen = screens[pref_screen_num]
        self.root = self.drawable = self.pref_screen.root

        # We don't probe the XFixes presence or version until we need it.
        self._xfixes_ready: bool | None = None

        # Probe the visuals (and related information), and make sure that our drawable is in an acceptable format.
        # These iterations and tests don't involve any traffic with the server; it's all stuff that was included in
        # the connection setup.  Effectively all modern setups will be acceptable, but we verify to be sure.

        # Currently, we assume that the drawable we're capturing is the root; when we add single-window capture,
        # we'll have to ask the server for its depth and visual.
        assert self.root == self.drawable  # noqa: S101
        self.drawable_depth = self.pref_screen.root_depth
        self.drawable_visual_id = self.pref_screen.root_visual
        # Server image byte order
        if xcb_setup.image_byte_order != xcb.ImageOrder.LSBFirst:
            msg = "Only X11 servers using LSB-First images are supported."
            raise ScreenShotError(msg)
        # Depth
        if self.drawable_depth not in SUPPORTED_DEPTHS:
            msg = f"Only screens of color depth 24 or 32 are supported, not {self.drawable_depth}"
            raise ScreenShotError(msg)
        # Format (i.e., bpp, padding)
        for format_ in xcb.setup_pixmap_formats(xcb_setup):
            if format_.depth == self.drawable_depth:
                break
        else:
            msg = f"Internal error: drawable's depth {self.drawable_depth} not found in screen's supported formats"
            raise ScreenShotError(msg)
        drawable_format = format_
        if drawable_format.bits_per_pixel != SUPPORTED_BITS_PER_PIXEL:
            msg = (
                f"Only screens at 32 bpp (regardless of color depth) are supported; "
                f"got {drawable_format.bits_per_pixel} bpp"
            )
            raise ScreenShotError(msg)
        if drawable_format.scanline_pad != SUPPORTED_BITS_PER_PIXEL:
            # To clarify the padding: the scanline_pad is the multiple that the scanline gets padded to.  If there
            # is no padding, then it will be the same as one pixel's size.
            msg = "Screens with scanline padding are not supported"
            raise ScreenShotError(msg)
        # Visual, the interpretation of pixels (like indexed, grayscale, etc).  (Visuals are arranged by depth, so
        # we iterate over the depths first.)
        for xcb_depth in xcb.screen_allowed_depths(self.pref_screen):
            if xcb_depth.depth == self.drawable_depth:
                break
        else:
            msg = "Internal error: drawable's depth not found in screen's supported depths"
            raise ScreenShotError(msg)
        for visual_info in xcb.depth_visuals(xcb_depth):
            if visual_info.visual_id == self.drawable_visual_id:
                break
        else:
            msg = "Internal error: drawable's visual not found in screen's supported visuals"
            raise ScreenShotError(msg)
        if visual_info.class_ not in {xcb.VisualClass.TrueColor, xcb.VisualClass.DirectColor}:
            msg = "Only TrueColor and DirectColor visuals are supported"
            raise ScreenShotError(msg)
        if (
            visual_info.red_mask != SUPPORTED_RED_MASK
            or visual_info.green_mask != SUPPORTED_GREEN_MASK
            or visual_info.blue_mask != SUPPORTED_BLUE_MASK
        ):
            # There are two ways to phrase this layout: BGRx accounts for the byte order, while xRGB implies the
            # native word order.  Since we return the data as a byte array, we use the former.  By the time we get
            # to this point, we've already checked the endianness and depth, so this is pretty much never going to
            # happen anyway.
            msg = "Only visuals with BGRx ordering are supported"
            raise ScreenShotError(msg)

    def close(self) -> None:
        """Close the XCB connection."""
        if self.conn is not None:
            xcb.disconnect(self.conn)
        self.conn = None

    def monitors(self) -> Monitors:
        """Populate monitor geometry information.

        Detects and returns monitor rectangles. The first entry
        represents the entire X11 root screen; subsequent entries,
        when available, represent individual monitors reported by
        XRandR.
        """
        if self.conn is None:
            msg = "Cannot identify monitors while the connection is closed"
            raise ScreenShotError(msg)

        monitors = []

        monitors.append(self._root_monitor())

        randr_version = self._randr_get_version()
        if randr_version is None or randr_version < (1, 2):
            return monitors

        # XRandR terminology (very abridged, but enough for this code):
        # - X screen / framebuffer: the overall drawable area for this root.
        # - CRTC: a display controller that scans out a rectangular region of the X screen.  A CRTC with zero
        #   outputs is inactive.  A CRTC may drive multiple outputs in clone/mirroring mode.
        # - Output: a physical connector (e.g. "HDMI-1", "DP-1").  The RandR "connection" state (connected vs
        #   disconnected) is separate from whether the output is currently driven by a CRTC.
        # - Monitor (RandR 1.5+): a logical rectangle presented to clients.  Monitors may be client-defined (useful
        #   for tiled displays) and are the closest match to what MSS wants.
        #
        # This implementation prefers RandR 1.5+ Monitors when available; otherwise it falls back to enumerating
        # active CRTCs.

        primary_output = self._randr_get_primary_output(randr_version)
        edid_atom = self._randr_get_edid_atom()

        if randr_version >= (1, 5):
            monitors += self._monitors_from_randr_monitors(primary_output, edid_atom)
        else:
            monitors += self._monitors_from_randr_crtcs(randr_version, primary_output, edid_atom)

        return monitors

    def _root_monitor(self) -> Monitor:
        if self.conn is None:
            msg = "Cannot identify monitors while the connection is closed"
            raise ScreenShotError(msg)

        root_geom = xcb.get_geometry(self.conn, self.root)
        return {
            "left": root_geom.x,
            "top": root_geom.y,
            "width": root_geom.width,
            "height": root_geom.height,
        }

    def _randr_get_version(self) -> tuple[int, int] | None:
        if self.conn is None:
            msg = "Cannot identify monitors while the connection is closed"
            raise ScreenShotError(msg)

        randr_ext_data = xcb.get_extension_data(self.conn, LIB.randr_id)
        if not randr_ext_data.present:
            return None

        randr_version_data = xcb.randr_query_version(self.conn, xcb.RANDR_MAJOR_VERSION, xcb.RANDR_MINOR_VERSION)
        return (randr_version_data.major_version, randr_version_data.minor_version)

    def _randr_get_primary_output(self, randr_version: tuple[int, int], /) -> xcb.RandrOutput | None:
        if self.conn is None:
            msg = "Cannot identify monitors while the connection is closed"
            raise ScreenShotError(msg)

        if randr_version >= (1, 3):
            primary_output_data = xcb.randr_get_output_primary(self.conn, self.drawable)
            return primary_output_data.output
        # Python None means that there was no way to identify a primary output.  This is distinct from XCB_NONE (that
        # is, xcb.RandROutput(0)), which means that there is not a primary monitor.
        return None

    def _randr_get_edid_atom(self) -> xcb.Atom | None:
        if self.conn is None:
            msg = "Cannot identify monitors while the connection is closed"
            raise ScreenShotError(msg)

        edid_atom = xcb.intern_atom(self.conn, "EDID", only_if_exists=True)
        if edid_atom is not None:
            return edid_atom

        # Formerly, "EDID" was known as "EdidData".  I don't know when it changed.
        return xcb.intern_atom(self.conn, "EdidData", only_if_exists=True)

    def _randr_output_ids(
        self,
        output: xcb.RandrOutput,
        timestamp: xcb.Timestamp,
        edid_atom: xcb.Atom | None,
        /,
    ) -> dict[str, Any]:
        if self.conn is None:
            msg = "Cannot identify monitors while the connection is closed"
            raise ScreenShotError(msg)

        output_info = xcb.randr_get_output_info(self.conn, output, timestamp)
        if output_info.status != 0:
            msg = "Display configuration changed while detecting monitors."
            raise ScreenShotError(msg)

        rv: dict[str, Any] = {}

        output_name_arr = xcb.randr_get_output_info_name(output_info)
        rv["output"] = bytes(output_name_arr).decode("utf_8", errors="replace")

        if edid_atom is not None:
            edid_prop = xcb.randr_get_output_property(
                self.conn,  # connection
                output,  # output
                edid_atom,  # property
                xcb.XCB_NONE,  # property type: Any
                0,  # long-offset: 0
                1024,  # long-length: in 4-byte units; 4k is plenty for an EDID
                0,  # delete: false
                0,  # pending: false
            )
            if edid_prop.type_.value != 0:
                edid_block = bytes(xcb.randr_get_output_property_data(edid_prop))
                edid_data = parse_edid(edid_block)
                if (display_name := edid_data.get("display_name")) is not None:
                    rv["name"] = display_name

                edid_params: dict[str, str] = {}
                if (id_legacy := edid_data.get("id_legacy")) is not None:
                    edid_params["model"] = id_legacy
                if (serial_number := edid_data.get("serial_number")) is not None:
                    edid_params["serial"] = str(serial_number)
                if (manufacture_year := edid_data.get("manufacture_year")) is not None:
                    if (manufacture_week := edid_data.get("manufacture_week")) is not None:
                        edid_params["mfr_date"] = f"{manufacture_year:04d}W{manufacture_week:02d}"
                    else:
                        edid_params["mfr_date"] = f"{manufacture_year:04d}"
                if (model_year := edid_data.get("model_year")) is not None:
                    edid_params["model_year"] = f"{model_year:04d}"
                if edid_params:
                    rv["unique_id"] = urlencode(edid_params)

        return rv

    @staticmethod
    def _choose_randr_output(
        outputs: Array[xcb.RandrOutput], primary_output: xcb.RandrOutput | None, /
    ) -> xcb.RandrOutput:
        if len(outputs) == 0:
            msg = "No RandR outputs available"
            raise ScreenShotError(msg)
        if primary_output is None:
            # We don't want to use the `in` check if this could be None, according to MyPy.
            return outputs[0]
        return primary_output if primary_output in outputs else outputs[0]

    def _monitors_from_randr_monitors(
        self, primary_output: xcb.RandrOutput | None, edid_atom: xcb.Atom | None, /
    ) -> Monitors:
        if self.conn is None:
            msg = "Cannot identify monitors while the connection is closed"
            raise ScreenShotError(msg)

        monitors = []

        monitors_reply = xcb.randr_get_monitors(self.conn, self.drawable, 1)
        timestamp = monitors_reply.timestamp
        for randr_monitor in xcb.randr_get_monitors_monitors(monitors_reply):
            monitor = {
                "left": randr_monitor.x,
                "top": randr_monitor.y,
                "width": randr_monitor.width,
                "height": randr_monitor.height,
                # Under XRandR, it's legal for no monitor to be primary.  In this case, case MSSBase.primary_monitor
                # will return the first monitor.  That said, we note in the dict that we explicitly are told by XRandR
                # that all of the monitors are not primary.  (This is distinct from the XRandR 1.2 path, which doesn't
                # have any information about primary monitors.)
                "is_primary": bool(randr_monitor.primary),
            }

            if randr_monitor.nOutput > 0:
                outputs = xcb.randr_monitor_info_outputs(randr_monitor)
                chosen_output = self._choose_randr_output(outputs, primary_output)
                monitor |= self._randr_output_ids(chosen_output, timestamp, edid_atom)

            monitors.append(monitor)

        return monitors

    def _monitors_from_randr_crtcs(
        self,
        randr_version: tuple[int, int],
        primary_output: xcb.RandrOutput | None,
        edid_atom: xcb.Atom | None,
        /,
    ) -> Monitors:
        if self.conn is None:
            msg = "Cannot identify monitors while the connection is closed"
            raise ScreenShotError(msg)

        monitors = []

        screen_resources: xcb.RandrGetScreenResourcesReply | xcb.RandrGetScreenResourcesCurrentReply
        if hasattr(LIB.randr, "xcb_randr_get_screen_resources_current") and randr_version >= (1, 3):
            screen_resources = xcb.randr_get_screen_resources_current(self.conn, self.drawable)
            crtcs = xcb.randr_get_screen_resources_current_crtcs(screen_resources)
        else:
            screen_resources = xcb.randr_get_screen_resources(self.conn, self.drawable)
            crtcs = xcb.randr_get_screen_resources_crtcs(screen_resources)
        timestamp = screen_resources.config_timestamp

        for crtc in crtcs:
            crtc_info = xcb.randr_get_crtc_info(self.conn, crtc, timestamp)
            if crtc_info.num_outputs == 0:
                continue
            monitor = {
                "left": crtc_info.x,
                "top": crtc_info.y,
                "width": crtc_info.width,
                "height": crtc_info.height,
            }

            outputs = xcb.randr_get_crtc_info_outputs(crtc_info)
            chosen_output = self._choose_randr_output(outputs, primary_output)
            monitor |= self._randr_output_ids(chosen_output, timestamp, edid_atom)
            # The concept of primary outputs was added in XRandR 1.3.  We distinguish between "all the monitors are
            # not primary" (RRGetOutputPrimary returned XCB_NONE, a valid case) and "we have no way to get
            # information about the primary monitor": in the latter case, we don't populate "is_primary".
            if primary_output is not None:
                monitor["is_primary"] = chosen_output == primary_output

            monitors.append(monitor)

        return monitors

    def _cursor_check_xfixes(self) -> bool:
        """Check XFixes availability and version.

        :returns: ``True`` if the server provides XFixes with a compatible
            version, otherwise ``False``.
        """
        if self.conn is None:
            msg = "Cannot take screenshot while the connection is closed"
            raise ScreenShotError(msg)

        xfixes_ext_data = xcb.get_extension_data(self.conn, LIB.xfixes_id)
        if not xfixes_ext_data.present:
            return False

        reply = xcb.xfixes_query_version(self.conn, xcb.XFIXES_MAJOR_VERSION, xcb.XFIXES_MINOR_VERSION)
        # We can work with 2.0 and later, but not sure about the actual minimum version we can use.  That's ok;
        # everything these days is much more modern.
        return (reply.major_version, reply.minor_version) >= (2, 0)

    def cursor(self) -> ScreenShot:
        """Capture the current cursor image.

        Pixels are returned in BGRA ordering.

        :returns: A screenshot object containing the cursor image and region.
        """

        if self.conn is None:
            msg = "Cannot take screenshot while the connection is closed"
            raise ScreenShotError(msg)

        if self._xfixes_ready is None:
            self._xfixes_ready = self._cursor_check_xfixes()
        if not self._xfixes_ready:
            msg = "Server does not have XFixes, or the version is too old."
            raise ScreenShotError(msg)

        cursor_img = xcb.xfixes_get_cursor_image(self.conn)
        region = {
            "left": cursor_img.x - cursor_img.xhot,
            "top": cursor_img.y - cursor_img.yhot,
            "width": cursor_img.width,
            "height": cursor_img.height,
        }

        data_arr = xcb.xfixes_get_cursor_image_cursor_image(cursor_img)
        data = bytearray(data_arr)
        # We don't need to do the same array slice-and-dice work as the Xlib-based implementation: Xlib has an
        # unfortunate historical accident that makes it have to return the cursor image in a different format.

        return ScreenShot(data, region)

    def _grab_xgetimage(self, monitor: Monitor, /) -> bytearray:
        """Retrieve pixels from a monitor using ``GetImage``.

        Used by the XGetImage backend and by the XShmGetImage backend in
        fallback mode.

        :param monitor: Monitor rectangle specifying ``left``, ``top``,
            ``width``, and ``height`` to capture.
        :returns: A screenshot object containing the captured region.
        """

        if self.conn is None:
            msg = "Cannot take screenshot while the connection is closed"
            raise ScreenShotError(msg)

        img_reply = xcb.get_image(
            self.conn,
            xcb.ImageFormat.ZPixmap,
            self.drawable,
            monitor["left"],
            monitor["top"],
            monitor["width"],
            monitor["height"],
            ALL_PLANES,
        )

        # Now, save the image.  This is a reference into the img_reply structure.
        img_data_arr = xcb.get_image_data(img_reply)
        # Copy this into a new bytearray, so that it will persist after we clear the image structure.
        img_data = bytearray(img_data_arr)

        if img_reply.depth != self.drawable_depth or img_reply.visual != self.drawable_visual_id:
            # This should never happen; a window can't change its visual.
            msg = (
                "Server returned an image with a depth or visual different than it initially reported: "
                f"expected {self.drawable_depth},{hex(self.drawable_visual_id.value)}, "
                f"got {img_reply.depth},{hex(img_reply.visual.value)}"
            )
            raise ScreenShotError(msg)

        return img_data



# This is part of the MSS Python's module.
# Source: https://github.com/BoboTiG/python-mss.

from __future__ import annotations

import platform
import warnings
from abc import ABC, abstractmethod
from datetime import datetime
from threading import Lock
from typing import TYPE_CHECKING, Any

from mss.exception import ScreenShotError
from mss.screenshot import ScreenShot
from mss.tools import to_png

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from mss.models import Monitor, Monitors, Size

    # Prior to 3.11, Python didn't have the Self type.  typing_extensions does, but we don't want to depend on it.
    try:
        from typing import Self
    except ImportError:
        try:
            from typing_extensions import Self
        except ImportError:
            Self = Any  # type: ignore[assignment]

try:
    from datetime import UTC
except ImportError:
    # Python < 3.11
    from datetime import timezone

    UTC = timezone.utc

#: Global lock protecting access to platform screenshot calls.
#:
#: .. versionadded:: 6.0.0
#:
#: .. deprecated:: 10.2.0
#:    The global lock is no longer used, and will be removed in a future release.
#:    MSS objects now have their own locks, which are not publicly-accessible.
lock = Lock()

OPAQUE = 255


# A sentinel value to indicate that a parameter was not passed, as opposed to being passed with a value of None.  This
# is used in the MSS constructor to distinguish between the user not passing a parameter, and the user explicitly
# passing None (which is the default for some parameters).  This allows us to preserve the existing behavior of ignoring
# certain parameters on certain platforms, while still allowing users to explicitly set those parameters on platforms
# where they are supported.
class _PlatformSpecific:
    def __init__(self, sphinx_repr: Any) -> None:
        self.sphinx_repr = str(sphinx_repr)

    def __repr__(self) -> str:
        # This is used to get Sphinx to show a useful default when it shows the default in the summary, rather than
        # an opaque object.
        return self.sphinx_repr


__all__ = ()


class MSSImplementation(ABC):
    """Base class for internal platform/backend implementations.

    Only one of these methods will be called at a time; the containing
    MSS object will hold a lock during these calls.
    """

    __slots__ = ("performance_status", "with_cursor")

    with_cursor: bool

    def __init__(self, /, *, with_cursor: bool = False) -> None:
        # We put with_cursor on the MSSImplementation because the Xlib backend will turn it off if the library isn't
        # installed.  (It's not a separate library under XCB.)  So, we need to let the backend mutate it.  Note that
        # the other platforms don't support with_cursor, and don't pass it to us.
        #
        # TODO(jholveck): #493 We should remove this expectation in 11.0.  It seems unlikely to be practically useful,
        # Xlib is legacy, and just complicates things.
        self.with_cursor = with_cursor

        # Any notes the backend needs to give the user for debugging purposes, like why it had to fall back to a
        # slower implementation.
        self.performance_status: list[str] = []

    @abstractmethod
    def cursor(self) -> ScreenShot | None:
        """Retrieve all cursor data. Pixels have to be RGB."""

    @abstractmethod
    def grab(self, monitor: Monitor, /) -> bytearray | tuple[bytearray, Size]:
        """Retrieve all pixels from a monitor. Pixels have to be RGB.

        If the monitor size is not in pixel units, include a Size in
        pixels (see issue #23).
        """

    @abstractmethod
    def monitors(self) -> Monitors:
        """Return positions of monitors."""

    def close(self) -> None:  # noqa: B027 - intentionally empty
        """Clean up.

        This will be called at most once.

        It's not necessary for subclasses to implement this if they
        have nothing to clean up.
        """

    @staticmethod
    def _cfactory(
        attr: Any,
        func: str,
        argtypes: list[Any],
        restype: Any,
        /,
        errcheck: Callable | None = None,
    ) -> None:
        """Factory to create a ctypes function and automatically manage errors."""
        meth = getattr(attr, func)
        meth.argtypes = argtypes
        meth.restype = restype
        if errcheck:
            meth.errcheck = errcheck


def _choose_impl(**kwargs: Any) -> MSSImplementation:
    """Return the backend implementation for the current platform.

    Detects the platform we are running on and instantiates the
    appropriate internal implementation class.

    .. seealso::
        - :class:`mss.MSS`
        - :class:`mss.darwin.MSS`
        - :class:`mss.linux.MSS`
        - :class:`mss.windows.MSS`
    """
    os_ = platform.system().lower()

    if os_ == "darwin":
        from mss.darwin import MSSImplDarwin  # noqa: PLC0415

        return MSSImplDarwin(**kwargs)

    if os_ == "linux":
        from mss.linux import choose_impl as choose_impl_linux  # noqa: PLC0415

        # Linux has its own factory to choose the backend.
        return choose_impl_linux(**kwargs)

    if os_ == "windows":
        from mss.windows import choose_impl as choose_impl_windows  # noqa: PLC0415

        # Windows has its own factory to choose the backend.
        return choose_impl_windows(**kwargs)

    msg = f"System {os_!r} not (yet?) implemented."
    raise ScreenShotError(msg)


# Does this belong here?
class MSS:
    """Multiple ScreenShots class

    :param backend: Backend selector, for platforms with multiple backends.
    :param compression_level: PNG compression level.
    :param with_cursor: Include the mouse cursor in screenshots
        (GNU/Linux only)
    :type display: bool, optional (default False)
    :param display: X11 display name (GNU/Linux only).
    :type display: bytes | str, optional (default :envvar:`$DISPLAY`)
    :param max_displays: Maximum number of displays to enumerate (macOS only).
    :type max_displays: int, optional (default 32)

    .. versionadded:: 8.0.0
        ``compression_level``, ``display``, ``max_displays``, and
        ``with_cursor`` keyword arguments.

    .. versionadded:: 10.2.0
        ``backend`` keyword argument.
    """

    # We want to:
    # * Let Sphinx, IDEs, and code-checkers know all the possible kwargs.
    # * Know if a user explicitly passed an unsupported platform-dependent keyword, so we can warn.
    # * Show a meaningful default in the Sphinx doc's summary string
    #
    # To accomplish this:
    # * We list the possibilities explicitly in the __init__ kwargs.
    # * We use a sentinel value, so we can tell whether or not the user actually gave us a value.
    # * We represent the "default value" sentinel object with something different, so Sphinx formats it usefully.
    _PD_WITH_CURSOR = _PlatformSpecific(False)  # noqa: FBT003
    _PD_DISPLAY = _PlatformSpecific(None)
    _PD_MAX_DISPLAYS = _PlatformSpecific(32)

    def __init__(
        self,
        /,
        *,
        backend: str = "default",
        compression_level: int = 6,
        with_cursor: bool | _PlatformSpecific = _PD_WITH_CURSOR,
        display: bytes | str | None | _PlatformSpecific = _PD_DISPLAY,
        max_displays: int | _PlatformSpecific = _PD_MAX_DISPLAYS,
    ) -> None:
        impl_kwargs = {}

        system = platform.system()
        for name, value, supported_platforms in [
            ("with_cursor", with_cursor, ["Linux"]),
            ("display", display, ["Linux"]),
            ("max_displays", max_displays, ["Darwin"]),
        ]:
            if isinstance(value, _PlatformSpecific):
                continue
            if system not in supported_platforms:
                # TODO(jholveck): #493 Accept platform-specific kwargs on all platforms for migration ease.  Foreign
                # kwargs are silently stripped with a warning.
                warnings.warn(
                    f"{name} is only available on {', '.join(supported_platforms)}. "
                    "This will be an error in the future.",
                    DeprecationWarning,
                    stacklevel=2,
                )
            else:
                impl_kwargs[name] = value

        self._impl: MSSImplementation = _choose_impl(
            backend=backend,
            **impl_kwargs,
        )

        # The cls_image is only used atomically, so does not require locking.
        self.cls_image: type[ScreenShot] = ScreenShot
        # The compression level is only used atomically, so does not require locking.
        #: PNG compression level used when saving the screenshot data into a file
        #: (see :py:func:`zlib.compress()` for details).
        #:
        #: .. versionadded:: 3.2.0
        self.compression_level = compression_level

        # The attributes below are protected by self._lock.  The attributes above are user-visible, so we don't
        # control when they're modified.  Currently, we only make sure that they're safe to modify while locked, or
        # document that the user shouldn't change them.  We could also use properties protect them against changes, or
        # change them under the lock.
        self._lock = Lock()
        self._monitors: Monitors | None = None
        self._closed = False

    def __enter__(self) -> Self:
        """For the cool call `with MSS() as mss:`."""
        return self

    def __exit__(self, *_: object) -> None:
        """For the cool call `with MSS() as mss:`."""
        self.close()

    def close(self) -> None:
        """Clean up.

        This releases resources that MSS may be using.  Once the MSS
        object is closed, it may not be used again.

        It is safe to call this multiple times; multiple calls have no
        effect.

        Rather than use :py:meth:`close` explicitly, we recommend you
        use the MSS object as a context manager::

            with mss.MSS() as sct:
                ...
        """
        with self._lock:
            if self._closed:
                return
            self._impl.close()
            self._closed = True

    def grab(self, monitor: Monitor | tuple[int, int, int, int], /) -> ScreenShot:
        """Retrieve screen pixels for a given monitor.

        Note: ``monitor`` can be a tuple like the one
        :py:meth:`PIL.ImageGrab.grab` accepts: ``(left, top, right, bottom)``

        :param monitor: The coordinates and size of the box to capture.
                        See :meth:`monitors <monitors>` for object details.
        :returns: Screenshot of the requested region.
        """
        # Convert PIL bbox style
        if isinstance(monitor, tuple):
            monitor = {
                "left": monitor[0],
                "top": monitor[1],
                "width": monitor[2] - monitor[0],
                "height": monitor[3] - monitor[1],
            }

        if monitor["width"] <= 0 or monitor["height"] <= 0:
            msg = f"Region has zero or negative size: {monitor!r}"
            raise ScreenShotError(msg)

        with self._lock:
            img_data_and_maybe_size = self._impl.grab(monitor)
            if isinstance(img_data_and_maybe_size, tuple):
                img_data, size = img_data_and_maybe_size
                screenshot = self.cls_image(img_data, monitor, size=size)
            else:
                img_data = img_data_and_maybe_size
                screenshot = self.cls_image(img_data, monitor)
            if self._impl.with_cursor and (cursor := self._impl.cursor()):
                return self._merge(screenshot, cursor)
            return screenshot

    @property
    def monitors(self) -> Monitors:
        """Get positions of all monitors.
        If the monitor has rotation, you have to deal with it
        inside this method.

        This method has to fill ``self._monitors`` with all information
        and use it as a cache:

        - ``self._monitors[0]`` is a dict of all monitors together
        - ``self._monitors[N]`` is a dict of the monitor N (with N > 0)

        Each monitor is a dict with:

        - ``left``: the x-coordinate of the upper-left corner
        - ``top``: the y-coordinate of the upper-left corner
        - ``width``: the width
        - ``height``: the height
        - ``is_primary``: (optional) true if this is the primary monitor
        - ``name``: (optional) human-readable device name
        - ``unique_id``: (optional) platform-specific stable identifier for the monitor
        - ``output``: (optional, Linux only) monitor output name, compatible with xrandr
        """
        with self._lock:
            if self._monitors is None:
                self._monitors = self._impl.monitors()
                assert self._monitors is not None  # noqa: S101
            return self._monitors

    @property
    def primary_monitor(self) -> Monitor:
        """Get the primary monitor.

        Returns the monitor marked as primary. If no monitor is marked as primary
        (or the platform doesn't support primary monitor detection), returns the
        first monitor (at index 1).

        :raises ScreenShotError: If no monitors are available.

        .. versionadded:: 10.2.0
        """
        monitors = self.monitors
        if len(monitors) <= 1:  # Only the "all monitors" entry or empty
            msg = "No monitor found."
            raise ScreenShotError(msg)

        return next(
            (
                monitor
                for monitor in monitors[1:]  # Skip the "all monitors" entry at index 0
                if monitor.get("is_primary", False)
            ),
            monitors[1],  # Fallback to the first monitor if no primary is found
        )

    def save(
        self,
        /,
        *,
        mon: int = 0,
        output: str = "monitor-{mon}.png",
        callback: Callable[[str], None] | None = None,
    ) -> Iterator[str]:
        """Grab a screenshot and save it to a file.

        :param int mon: The monitor to screenshot (default=0). ``-1`` grabs all
            monitors, ``0`` grabs each monitor, and ``N`` grabs monitor ``N``.
        :param str output: The output filename. Keywords: ``{mon}``, ``{top}``,
            ``{left}``, ``{width}``, ``{height}``, ``{date}``.
        :param callable callback: Called before saving the screenshot; receives
            the ``output`` argument.
        :return: Created file(s).
        """
        monitors = self.monitors
        if not monitors:
            msg = "No monitor found."
            raise ScreenShotError(msg)

        if mon == 0:
            # One screenshot by monitor
            for idx, monitor in enumerate(monitors[1:], 1):
                fname = output.format(mon=idx, date=datetime.now(UTC) if "{date" in output else None, **monitor)
                if callable(callback):
                    callback(fname)
                sct = self.grab(monitor)
                to_png(sct.rgb, sct.size, level=self.compression_level, output=fname)
                yield fname
        else:
            # A screenshot of all monitors together or
            # a screenshot of the monitor N.
            mon = 0 if mon == -1 else mon
            try:
                monitor = monitors[mon]
            except IndexError as exc:
                msg = f"Monitor {mon!r} does not exist."
                raise ScreenShotError(msg) from exc

            output = output.format(mon=mon, date=datetime.now(UTC) if "{date" in output else None, **monitor)
            if callable(callback):
                callback(output)
            sct = self.grab(monitor)
            to_png(sct.rgb, sct.size, level=self.compression_level, output=output)
            yield output

    def shot(self, /, **kwargs: Any) -> str:
        """Helper to save the screenshot of the 1st monitor, by default.
        You can pass the same arguments as for :meth:`save`.
        """
        kwargs["mon"] = kwargs.get("mon", 1)
        return next(self.save(**kwargs))

    @staticmethod
    def _merge(screenshot: ScreenShot, cursor: ScreenShot, /) -> ScreenShot:
        """Create composite image by blending screenshot and mouse cursor.

        The cursor image should be in straight (not premultiplied) alpha.
        """
        (cx, cy), (cw, ch) = cursor.pos, cursor.size
        (x, y), (w, h) = screenshot.pos, screenshot.size

        cx2, cy2 = cx + cw, cy + ch
        x2, y2 = x + w, y + h

        overlap = cx < x2 and cx2 > x and cy < y2 and cy2 > y
        if not overlap:
            return screenshot

        screen_raw = screenshot.raw
        cursor_raw = cursor.raw

        cy, cy2 = (cy - y) * 4, (cy2 - y2) * 4
        cx, cx2 = (cx - x) * 4, (cx2 - x2) * 4
        start_count_y = -cy if cy < 0 else 0
        start_count_x = -cx if cx < 0 else 0
        stop_count_y = ch * 4 - max(cy2, 0)
        stop_count_x = cw * 4 - max(cx2, 0)
        rgb = range(3)

        for count_y in range(start_count_y, stop_count_y, 4):
            pos_s = (count_y + cy) * w + cx
            pos_c = count_y * cw

            for count_x in range(start_count_x, stop_count_x, 4):
                spos = pos_s + count_x
                cpos = pos_c + count_x
                alpha = cursor_raw[cpos + 3]

                if not alpha:
                    continue

                if alpha == OPAQUE:
                    screen_raw[spos : spos + 3] = cursor_raw[cpos : cpos + 3]
                else:
                    alpha2 = alpha / 255
                    for i in rgb:
                        screen_raw[spos + i] = int(cursor_raw[cpos + i] * alpha2 + screen_raw[spos + i] * (1 - alpha2))

        return screenshot

    @staticmethod
    def _cfactory(
        attr: Any,
        func: str,
        argtypes: list[Any],
        restype: Any,
        /,
        errcheck: Callable | None = None,
    ) -> None:
        """Factory to create a ctypes function and automatically manage errors."""
        meth = getattr(attr, func)
        meth.argtypes = argtypes
        meth.restype = restype
        if errcheck:
            meth.errcheck = errcheck

    # Some backends may expose additional read-only attributes.  Those
    # are implemented here, as properties.  By making them properties,
    # instead of using __getattr__, they're also accessible to Sphinx
    # and type checkers.
    #
    # Important: We need to be judicious in what we add here.  We
    # really don't want these to proliferate.  Some, like
    # max_displays, should probably be removed in 11.0.  with_cursor
    # should probably be moved to MSS instead of MSSImplementation (as
    # noted there).

    @property
    def performance_status(self) -> list[str]:
        """Implementation-specific notes that might affect performance.

        For instance, on GNU/Linux, when using the default XShmGetImage
        backend, this will include a note if the MIT-SHM extension is
        not usable.

        This may not be ready until one screenshot has been taken.

        This is meant only for debugging purposes; the contents are
        subject to change at any time.

        .. versionadded:: 10.2.0
        """
        return self._impl.performance_status

    @property
    def max_displays(self) -> int:
        """Maximum number of displays to handle.

        Availability: macOS

        .. versionadded:: 8.0.0
        """
        return self._impl.max_displays  # type: ignore[attr-defined]

    @property
    def with_cursor(self) -> bool:
        """Include the mouse cursor in screenshots.

        In some circumstances, it may not be possible to include the
        cursor.  In that case, MSS will automatically change this to
        False when the object is created.

        This cannot be changed after creating the object.

        .. versionadded:: 8.0.0
        """
        return self._impl.with_cursor


# TODO(jholveck): #493 Remove compatibility alias after 10.x transition period.
MSSBase = MSS



"""This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss.
"""

import os.path
import platform
import sys
from argparse import ArgumentError, ArgumentParser

from mss import MSS, __version__
from mss.exception import ScreenShotError
from mss.tools import to_png


def _backend_cli_choices() -> list[str]:
    os_name = platform.system().lower()
    if os_name == "darwin":
        from mss import darwin  # noqa: PLC0415

        return darwin.BACKENDS
    if os_name == "linux":
        from mss import linux  # noqa: PLC0415

        return linux.BACKENDS
    if os_name == "windows":
        from mss import windows  # noqa: PLC0415

        return windows.BACKENDS
    return ["default"]


def main(*args: str) -> int:  # noqa: PLR0912
    """Main logic."""
    backend_choices = _backend_cli_choices()

    cli_args = ArgumentParser(prog="mss", exit_on_error=False)
    cli_args.add_argument(
        "-c",
        "--coordinates",
        default="",
        type=str,
        help="the part of the screen to capture: top, left, width, height",
    )
    cli_args.add_argument(
        "-l",
        "--level",
        default=6,
        type=int,
        choices=list(range(10)),
        help="the PNG compression level",
    )
    cli_args.add_argument("-m", "--monitor", default=0, type=int, help="the monitor to screenshot")
    cli_args.add_argument("-o", "--output", default="monitor-{mon}.png", help="the output file name")
    cli_args.add_argument("--with-cursor", default=None, action="store_true", help="include the cursor")
    cli_args.add_argument(
        "-q",
        "--quiet",
        default=False,
        action="store_true",
        help="do not print created files",
    )
    cli_args.add_argument(
        "-b", "--backend", default="default", choices=backend_choices, help="platform-specific backend to use"
    )
    cli_args.add_argument("-v", "--version", action="version", version=__version__)

    try:
        options = cli_args.parse_args(args or None)
    except ArgumentError as e:
        # By default, parse_args will print and the error and exit.  We
        # return instead of exiting, to make unit testing easier.
        cli_args.print_usage(sys.stderr)
        print(f"{cli_args.prog}: error: {e}", file=sys.stderr)
        return 2
    grab_kwargs = {"mon": options.monitor, "output": options.output}
    if options.coordinates:
        try:
            top, left, width, height = options.coordinates.split(",")
        except ValueError:
            print("Coordinates syntax: top, left, width, height")
            return 2

        grab_kwargs["mon"] = {
            "top": int(top),
            "left": int(left),
            "width": int(width),
            "height": int(height),
        }
        if options.output == "monitor-{mon}.png":
            grab_kwargs["output"] = "sct-{top}x{left}_{width}x{height}.png"

    if options.with_cursor is not None and platform.system().lower() != "linux":
        if not options.quiet:
            print("[WARNING] --with-cursor is only supported on Linux; ignoring.", file=sys.stderr)
        options.with_cursor = None

    mss_kwargs = {"backend": options.backend}
    if options.with_cursor is not None:
        mss_kwargs["with_cursor"] = options.with_cursor

    try:
        with MSS(**mss_kwargs) as sct:
            if options.coordinates:
                output = grab_kwargs["output"].format(**grab_kwargs["mon"])
                sct_img = sct.grab(grab_kwargs["mon"])
                to_png(sct_img.rgb, sct_img.size, level=options.level, output=output)
                if not options.quiet:
                    print(os.path.realpath(output))
            else:
                for file_name in sct.save(**grab_kwargs):
                    if not options.quiet:
                        print(os.path.realpath(file_name))
            return 0
    except ScreenShotError:
        if options.quiet:
            return 1
        raise


if __name__ == "__main__":
    try:
        sys.exit(main())
    except ScreenShotError as exc:
        print("[ERROR]", exc)
        sys.exit(1)
